"""
Utilities for working with the
`Open Images <https://storage.googleapis.com/openimages/web/index.html>`
dataset.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import csv
import logging
import os
import random
import warnings

import pandas as pd

import eta.core.image as etai
import eta.core.serial as etas
import eta.core.utils as etau
import eta.core.web as etaw

import fiftyone as fo
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.utils.aws as foua
import fiftyone.utils.data as foud


logger = logging.getLogger(__name__)


class OpenImagesV6DatasetImporter(foud.LabeledImageDatasetImporter):
    """Base class for importing datasets in Open Images V6 format.

    See :class:`fiftyone.types.OpenImagesV6Dataset` for format details.

    Args:
        dataset_dir: the dataset directory
        label_types (None): a label type or list of label types to load. The
            supported values are
            ``("detections", "classifications", "relationships", "segmentations")``.
            By default, all label types are loaded
        classes (None): a string or list of strings specifying required classes
            to load. If provided, only samples containing at least one instance
            of a specified class will be loaded
        attrs (None): a string or list of strings specifying required
            relationship attributes to load. Only applicable when
            ``label_types`` includes "relationships". If provided, only samples
            containing at least one instance of a specified attribute will be
            loaded
        image_ids (None): an optional list of specific image IDs to load. Can
            be provided in any of the following formats:

            -   a list of ``<image-id>`` strings
            -   a list of ``<split>/<image-id>`` strings
            -   the path to a text (newline-separated), JSON, or CSV file
                containing the list of image IDs to load in either of the first
                two formats
        include_id (True): whether to load the Open Images ID for each sample
            along with the labels
        only_matching (False): whether to only load labels that match the
            ``classes`` or ``attrs`` requirements that you provide (True), or
            to load all labels for samples that match the requirements (False)
        load_hierarchy (True): whether to load the classes hiearchy and add it
            to the dataset's ``info`` dictionary
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to load. If
            ``label_types``, ``classes``, and/or ``attrs`` are also specified,
            first priority will be given to samples that contain all of the
            specified label types, classes, and/or attributes, followed by
            samples that contain at least one of the specified labels types or
            classes. The actual number of samples loaded may be less than this
            maximum value if the dataset does not contain sufficient samples
            matching your requirements. By default, all matching samples are
            loaded
    """

    def __init__(
        self,
        dataset_dir,
        label_types=None,
        classes=None,
        attrs=None,
        image_ids=None,
        include_id=True,
        only_matching=False,
        load_hierarchy=True,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        _label_types = _parse_label_types(label_types)
        if include_id:
            _label_types.append("open_images_id")

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.label_types = label_types
        self.classes = classes
        self.attrs = attrs
        self.image_ids = image_ids
        self.only_matching = only_matching
        self.load_hierarchy = load_hierarchy

        self._label_types = _label_types
        self._images_map = None
        self._info = None
        self._classes_map = None
        self._attrs_map = None
        self._cls_data = None
        self._det_data = None
        self._rel_data = None
        self._seg_data = None
        self._uuids = None
        self._iter_uuids = None

    def __iter__(self):
        self._iter_uuids = iter(self._uuids)
        return self

    def __len__(self):
        return len(self._uuids)

    def __next__(self):
        image_id = next(self._iter_uuids)

        image_path = self._images_map[image_id]

        label = {}

        if "classifications" in self._label_types:
            # Add labels
            pos_labels, neg_labels = _create_classifications(
                self._cls_data, image_id, self._classes_map
            )
            if pos_labels is not None:
                label["positive_labels"] = pos_labels

            if neg_labels is not None:
                label["negative_labels"] = neg_labels

        if "detections" in self._label_types:
            # Add detections
            detections = _create_detections(
                self._det_data, image_id, self._classes_map
            )
            if detections is not None:
                label["detections"] = detections

        if "segmentations" in self._label_types:
            # Add segmentations
            segmentations = _create_segmentations(
                self._seg_data,
                image_id,
                self._classes_map,
                self.dataset_dir,
            )
            if segmentations is not None:
                label["segmentations"] = segmentations

        if "relationships" in self._label_types:
            # Add relationships
            relationships = _create_relationships(
                self._rel_data, image_id, self._classes_map, self._attrs_map
            )
            if relationships is not None:
                label["relationships"] = relationships

        if "open_images_id" in self._label_types:
            label["open_images_id"] = image_id

        if self._has_scalar_labels:
            label = next(iter(label.values())) if label else None

        return image_path, None, label

    @property
    def has_dataset_info(self):
        return True

    @property
    def has_image_metadata(self):
        return False

    @property
    def _has_scalar_labels(self):
        return (
            len(self._label_types) == 1
            and self._label_types[0] != "classifications"
        )

    @property
    def label_cls(self):
        types = {
            "classifications": fol.Classifications,
            "detections": fol.Detections,
            "segmentations": fol.Detections,
            "relationships": fol.Detections,
            "open_images_id": fof.StringField,
        }

        if self._has_scalar_labels:
            return types[self._label_types[0]]

        return {k: v for k, v in types.items() if k in self._label_types}

    def setup(self):
        dataset_dir = self.dataset_dir
        seed = self.seed
        shuffle = self.shuffle
        max_samples = self.max_samples
        label_types = self.label_types
        classes = self.classes
        attrs = self.attrs
        image_ids = self.image_ids

        data_dir = os.path.join(self.dataset_dir, "data")

        images_map = {
            os.path.splitext(filename)[0]: os.path.join(data_dir, filename)
            for filename in etau.list_files(data_dir)
        }
        available_ids = list(images_map.keys())

        info = {}

        self._images_map = images_map
        self._info = info

        if not available_ids:
            self._uuids = []
            return

        if image_ids:
            image_ids = _parse_image_ids(image_ids, ignore_split=True)
            image_ids = list(set(image_ids) & set(available_ids))
        else:
            image_ids = available_ids

        (
            classes_map,
            all_classes,
            classes,
            oi_classes,
            attrs_map,
            all_attrs,
            attrs,
            oi_attrs,
            seg_classes,
            _,
        ) = _setup(
            dataset_dir,
            label_types=label_types,
            classes=classes,
            attrs=attrs,
            seed=seed,
            download=False,
        )

        (
            cls_data,
            det_data,
            rel_data,
            seg_data,
            all_label_ids,
            any_label_ids,
            _,
        ) = _get_all_label_data(
            dataset_dir,
            image_ids,
            label_types=label_types,
            classes=classes,
            oi_classes=oi_classes,
            attrs=attrs,
            oi_attrs=oi_attrs,
            seg_classes=seg_classes,
            ids_only=False,
            track_all_ids=max_samples is not None,
            only_matching=self.only_matching,
            download=False,
        )

        if max_samples is not None:
            if self.classes is None and self.attrs is None:
                # No requirements were provided, so always make all samples
                # available
                extra_ids = set(image_ids) - set(any_label_ids)
            else:
                extra_ids = set()

            # Prioritize samples with all labels, then any, then extras
            not_all_ids = any_label_ids - all_label_ids
            all_label_ids = sorted(all_label_ids)
            not_all_ids = sorted(not_all_ids)
            extra_ids = sorted(extra_ids)

            if shuffle:
                random.shuffle(all_label_ids)
                random.shuffle(not_all_ids)
                random.shuffle(extra_ids)

            valid_ids = all_label_ids + not_all_ids + extra_ids
            valid_ids = valid_ids[:max_samples]
        else:
            if self.classes is None and self.attrs is None:
                # No requirements were provided, so always make all samples
                # available
                valid_ids = sorted(image_ids)
            else:
                valid_ids = sorted(any_label_ids)

            if shuffle:
                random.shuffle(valid_ids)

        if self.load_hierarchy:
            hierarchy, _ = _get_hierarchy(dataset_dir, download=False)
            info["hierarchy"] = hierarchy

        if attrs_map:
            info["attributes_map"] = attrs_map

        if all_attrs:
            info["attributes"] = all_attrs

        if seg_classes:
            info["segmentation_classes"] = seg_classes

        info["classes_map"] = classes_map
        info["classes"] = all_classes

        self._classes_map = classes_map
        self._attrs_map = attrs_map
        self._cls_data = cls_data
        self._det_data = det_data
        self._rel_data = rel_data
        self._seg_data = seg_data
        self._uuids = valid_ids

    def get_dataset_info(self):
        return self._info


def get_attributes(version="v6", dataset_dir=None):
    """Gets the list of relationship attributes in the Open Images dataset.

    Args:
        version ("v6"): the version of the Open Images dataset. Supported
            values are ``("v6")``
        dataset_dir (None): an optional root directory the in which the dataset
            is downloaded

    Returns:
        a sorted list of attribute names
    """
    _verify_version(version)

    if dataset_dir is None:
        dataset_dir = os.path.join(
            fo.config.dataset_zoo_dir, "open-images-%s" % version
        )

    try:
        # Try to use already downloaded file
        attrs_map, _ = _get_attrs_map(dataset_dir, download=False)
    except:
        # Download file to temporary location
        with etau.TempDir() as tmp_dir:
            attrs_map, _ = _get_attrs_map(tmp_dir, download=True)

    return sorted(attrs_map.values())


def get_classes(version="v6", dataset_dir=None):
    """Gets the boxable classes that exist in classifications, detections, and
    relationships in the Open Images dataset.

    This method can be called in isolation without downloading the dataset.

    Args:
        version ("v6"): the version of the Open Images dataset. Supported
            values are ``("v6")``
        dataset_dir (None): an optional root directory the in which the dataset
            is downloaded

    Returns:
        a sorted list of class name strings
    """
    _verify_version(version)

    if dataset_dir is None:
        dataset_dir = os.path.join(
            fo.config.dataset_zoo_dir, "open-images-%s" % version
        )

    try:
        # Try to use already downloaded file
        classes_map, _ = _get_classes_map(dataset_dir, download=False)
    except:
        # Download file to temporary location
        with etau.TempDir() as tmp_dir:
            classes_map, _ = _get_classes_map(tmp_dir, download=True)

    return sorted(classes_map.values())


def get_segmentation_classes(version="v6", dataset_dir=None):
    """Gets the list of classes (350) that are labeled with segmentations in
    the Open Images V6 dataset.

    This method can be called in isolation without downloading the dataset.

    Args:
        version ("v6"): the version of the Open Images dataset. Supported
            values are ``("v6")``
        dataset_dir (None): an optional root directory the in which the dataset
            is downloaded

    Returns:
        a sorted list of segmentation class name strings
    """
    _verify_version(version)

    if dataset_dir is None:
        dataset_dir = os.path.join(
            fo.config.dataset_zoo_dir, "open-images-%s" % version
        )

    try:
        # Try to use already downloaded file
        seg_classes, _ = _get_seg_classes(dataset_dir, download=False)
    except:
        # Download file to temporary location
        with etau.TempDir() as tmp_dir:
            seg_classes, _ = _get_seg_classes(tmp_dir, download=True)

    return seg_classes


def download_open_images_split(
    dataset_dir,
    split,
    version="v6",
    label_types=None,
    classes=None,
    attrs=None,
    image_ids=None,
    num_workers=None,
    shuffle=None,
    seed=None,
    max_samples=None,
):
    """Utility that downloads full or partial splits of the
    `Open Images dataset <https://storage.googleapis.com/openimages/web/index.html>`_.

    See :class:`fiftyone.types.OpenImagesV6Dataset` for the format in which
    ``dataset_dir`` will be arranged.

    Any existing files are not re-downloaded.

    This method specifically downloads the subsets of annotations corresponding
    to the 600 boxable classes of Open Images.
    `See here <https://storage.googleapis.com/openimages/web/download.html>`_
    for other download options.

    Args:
        dataset_dir: the directory to download the dataset
        split: the split to download. Supported values are
            ``("train", "validation", "test")``
        version ("v6"): the version of the Open Images dataset to download.
            Supported values are ``("v6")``
        label_types (None): a label type or list of label types to load. The
            supported values are
            ``("detections", "classifications", "relationships", "segmentations")``.
            By default, all label types are loaded
        classes (None): a string or list of strings specifying required classes
            to load. If provided, only samples containing at least one instance
            of a specified class will be loaded
        attrs (None): a string or list of strings specifying required
            relationship attributes to load. Only applicable when
            ``label_types`` includes "relationships". If provided, only samples
            containing at least one instance of a specified attribute will be
            loaded
        image_ids (None): an optional list of specific image IDs to load. Can
            be provided in any of the following formats:

            -   a list of ``<image-id>`` strings
            -   a list of ``<split>/<image-id>`` strings
            -   the path to a text (newline-separated), JSON, or CSV file
                containing the list of image IDs to load in either of the first
                two formats
        num_workers (None): the number of processes to use when downloading
            individual images. By default, ``multiprocessing.cpu_count()`` is
            used
        shuffle (False): whether to randomly shuffle the order in which samples
            are chosen for partial downloads
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to load per split. If
            ``label_types``, ``classes``, and/or ``attrs`` are also specified,
            first priority will be given to samples that contain all of the
            specified label types, classes, and/or attributes, followed by
            samples that contain at least one of the specified labels types or
            classes. The actual number of samples loaded may be less than this
            maximum value if the dataset does not contain sufficient samples
            matching your requirements. By default, all matching samples are
            loaded

    Returns:
        a tuple of:

        -   num_samples: the total number of downloaded images, or ``None`` if
            everything was already downloaded
        -   classes: the list of all classes, or ``None`` if everything was
            already downloaded
        -   did_download: whether any content was downloaded (True) or if all
            necessary files were already downloaded (False)
    """
    _verify_version(version)

    did_download = False

    if image_ids is not None:
        image_ids, _did_download = _parse_and_verify_image_ids(
            image_ids, dataset_dir, split, download=True
        )
        did_download |= _did_download
    else:
        image_ids, _did_download = _load_all_image_ids(
            dataset_dir, split=split, download=True
        )
        did_download |= _did_download

    downloaded_ids = _get_downloaded_image_ids(dataset_dir)

    (
        classes_map,
        all_classes,
        classes,
        oi_classes,
        attrs_map,
        all_attrs,
        attrs,
        oi_attrs,
        seg_classes,
        _did_download,
    ) = _setup(
        dataset_dir,
        label_types=label_types,
        classes=classes,
        attrs=attrs,
        seed=seed,
        download=True,
    )
    did_download |= _did_download

    # Download class hierarchy if necessary (used in evaluation)
    _, _did_download = _get_hierarchy(
        dataset_dir, classes_map=classes_map, download=True
    )
    did_download |= _did_download

    num_samples, _did_download = _download(
        image_ids,
        downloaded_ids,
        oi_classes,
        oi_attrs,
        seg_classes,
        dataset_dir,
        split,
        label_types=label_types,
        classes=classes,
        attrs=attrs,
        max_samples=max_samples,
        shuffle=shuffle,
        num_workers=num_workers,
        download=True,
    )
    did_download |= _did_download

    return num_samples, all_classes, did_download


def _setup(
    dataset_dir,
    label_types=None,
    classes=None,
    attrs=None,
    seed=None,
    download=False,
):
    did_download = False
    _label_types = _parse_label_types(label_types)

    if etau.is_str(classes):
        classes = [classes]

    if etau.is_str(attrs):
        attrs = [attrs]

    if seed is not None:
        random.seed(seed)

    # Map of class IDs to class names
    classes_map, _did_download = _get_classes_map(
        dataset_dir, download=download
    )
    classes_map_rev = {v: k for k, v in classes_map.items()}
    did_download |= _did_download

    all_classes = sorted(classes_map.values())

    if classes is not None:
        oi_classes = []
        missing_classes = []
        filtered_classes = []
        for c in classes:
            if c in classes_map_rev:
                oi_classes.append(classes_map_rev[c])
                filtered_classes.append(c)
            else:
                missing_classes.append(c)

        classes = filtered_classes
        if missing_classes:
            logger.warning(
                "Ignoring invalid classes %s\nYou can view the available "
                "classes via `fiftyone.utils.openimages.get_classes()`",
                missing_classes,
            )
    else:
        oi_classes = None

    if "relationships" in _label_types:
        # Map of attribute IDs to attribute names
        attrs_map, _did_download = _get_attrs_map(
            dataset_dir, download=download
        )
        attrs_map_rev = {v: k for k, v in attrs_map.items()}
        did_download |= _did_download

        all_attrs = sorted(attrs_map.values())

        if attrs is None:
            oi_attrs = [attrs_map_rev[a] for a in all_attrs]
        else:
            oi_attrs = []
            missing_attrs = []
            filtered_attrs = []
            for a in attrs:
                if a in attrs_map_rev:
                    oi_attrs.append(attrs_map_rev[a])
                    filtered_attrs.append(a)
                else:
                    missing_attrs.append(a)

            attrs = filtered_attrs
            if missing_attrs:
                logger.warning(
                    "Ignoring invalid attributes %s\nYou can view the "
                    "available attributes via "
                    "`fiftyone.utils.openimages.get_attributes()`",
                    missing_attrs,
                )
    else:
        attrs = None
        attrs_map = None
        oi_attrs = None
        all_attrs = None

    if "segmentations" in _label_types:
        seg_classes, _did_download = _get_seg_classes(
            dataset_dir, classes_map=classes_map, download=download
        )
        did_download |= _did_download
    else:
        seg_classes = None

    return (
        classes_map,
        all_classes,
        classes,
        oi_classes,
        attrs_map,
        all_attrs,
        attrs,
        oi_attrs,
        seg_classes,
        did_download,
    )


def _get_general_metadata_file(dataset_dir, filename, url, download=True):
    filepath = os.path.join(dataset_dir, "metadata", filename)
    if not os.path.exists(filepath):
        for split in _SUPPORTED_SPLITS:
            split_filepath = os.path.join(
                dataset_dir, split, "metadata", filename
            )
            if os.path.exists(split_filepath):
                return split_filepath, False

    did_download = _download_file_if_necessary(
        filepath, url, quiet=0, download=download
    )

    return filepath, did_download


def _get_attrs_map(dataset_dir, download=True):
    url = _ANNOTATION_DOWNLOAD_URLS["general"]["attr_names"]
    attrs_csv, did_download = _get_general_metadata_file(
        dataset_dir, "attributes.csv", url, download=download
    )
    attrs_data = _parse_csv(attrs_csv)
    attrs_map = {k: v for k, v in attrs_data}
    return attrs_map, did_download


def _get_classes_map(dataset_dir, download=True):
    url = _ANNOTATION_DOWNLOAD_URLS["general"]["class_names"]
    cls_csv, did_download = _get_general_metadata_file(
        dataset_dir, "classes.csv", url, download=download
    )
    cls_data = _parse_csv(cls_csv)
    classes_map = {k: v for k, v in cls_data}
    return classes_map, did_download


def _get_seg_classes(dataset_dir, classes_map=None, download=True):
    did_download = False

    if not classes_map:
        classes_map, _did_download = _get_classes_map(
            dataset_dir, download=download
        )
        did_download |= _did_download

    url = _ANNOTATION_DOWNLOAD_URLS["general"]["segmentation_classes"]
    seg_cls_txt, _did_download = _get_general_metadata_file(
        dataset_dir, "segmentation_classes.csv", url, download=download
    )
    did_download |= _did_download

    with open(seg_cls_txt, "r", encoding="utf8") as f:
        seg_classes_oi = [l.rstrip("\n") for l in f]

    seg_classes = [classes_map[c] for c in seg_classes_oi]

    return sorted(seg_classes), did_download


def _get_hierarchy(dataset_dir, classes_map=None, download=True):
    hierarchy_path = os.path.join(dataset_dir, "metadata", "hierarchy.json")

    if os.path.exists(hierarchy_path):
        hierarchy = etas.load_json(hierarchy_path)
        return hierarchy, False

    if not download:
        raise ValueError("Hierarchy file '%s' not found" % hierarchy_path)

    with etau.TempDir() as tmp_dir:
        url = _ANNOTATION_DOWNLOAD_URLS["general"]["hierarchy"]
        tmp_filepath, _ = _get_general_metadata_file(
            tmp_dir, "hierarchy.json", url, download=download
        )

        hierarchy = etas.load_json(tmp_filepath)

        if classes_map is None:
            classes_map, _ = _get_classes_map(tmp_dir, download=download)

        # Not included in standard classes
        entity_classes_map = {"/m/0bl9f": "Entity"}
        entity_classes_map.update(classes_map)
        hierarchy = _rename_subcategories(hierarchy, entity_classes_map)
        etas.write_json(hierarchy, hierarchy_path)

    return hierarchy, True


def _rename_subcategories(hierarchy, classes_map):
    if "LabelName" in hierarchy.keys():
        curr_label = hierarchy["LabelName"]
        hierarchy["LabelName"] = classes_map[curr_label]

    if "Subcategory" in hierarchy.keys():
        subs = []
        for sub in hierarchy["Subcategory"]:
            subs.append(_rename_subcategories(sub, classes_map))

        hierarchy["Subcategory"] = subs

    if "Part" in hierarchy.keys():
        parts = []
        for part in hierarchy["Part"]:
            parts.append(_rename_subcategories(part, classes_map))

        hierarchy["Part"] = parts

    return hierarchy


def _parse_csv(filename, dataframe=False, index_col=None):
    if dataframe:
        data = pd.read_csv(filename, index_col=index_col)
    else:
        with open(filename, "r", newline="", encoding="utf8") as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(10240))
            csvfile.seek(0)
            if dialect.delimiter in _CSV_DELIMITERS:
                reader = csv.reader(csvfile, dialect)
            else:
                reader = csv.reader(csvfile)

            data = [row for row in reader]

    return data


def _parse_image_ids(image_ids, ignore_split=False):
    # Load IDs from file
    if etau.is_str(image_ids):
        image_ids_path = image_ids
        ext = os.path.splitext(image_ids_path)[-1]
        if ext == ".txt":
            with open(image_ids_path, "r") as f:
                image_ids = [i for i in f.readlines()]

        elif ext == ".json":
            image_ids = etas.load_json(image_ids_path)

        elif ext == ".csv":
            image_ids = _parse_csv(image_ids_path)
            if isinstance(image_ids[0], list):
                # Flatten list
                image_ids = [i for lst in image_ids for i in lst]

        else:
            raise ValueError(
                "Invalid image ID file '%s'. Supported formats are .txt, "
                ".csv, and .json" % ext
            )

    image_ids = [i.strip() for i in image_ids]

    if ignore_split:
        # Remove `split/` prefix, if any
        image_ids = [os.path.basename(i) for i in image_ids]

    return image_ids


def _parse_and_verify_image_ids(image_ids, dataset_dir, split, download=True):
    image_ids = _parse_image_ids(image_ids)

    split_image_ids = []
    unspecified_split_ids = []

    # Ignore images from other splits
    for image_id in image_ids:
        if "/" in image_id:
            _split, image_id = image_id.split("/")
            if _split not in _SUPPORTED_SPLITS:
                raise ValueError(
                    "Invalid split '%s'. Supported splits are %s"
                    % (_split, _SUPPORTED_SPLITS)
                )

            if _split != split:
                continue

            split_image_ids.append(image_id)
        else:
            unspecified_split_ids.append(image_id)

    return _verify_image_ids(
        split_image_ids,
        unspecified_split_ids,
        dataset_dir,
        split,
        download=download,
    )


def _parse_label_types(label_types):
    if label_types is None:
        return _SUPPORTED_LABEL_TYPES

    if etau.is_str(label_types):
        label_types = [label_types]
    else:
        label_types = list(label_types)

    bad_types = [l for l in label_types if l not in _SUPPORTED_LABEL_TYPES]

    if len(bad_types) == 1:
        raise ValueError(
            "Unsupported label type '%s'. Supported types are %s"
            % (bad_types[0], _SUPPORTED_LABEL_TYPES)
        )

    if len(bad_types) > 1:
        raise ValueError(
            "Unsupported label types %s. Supported types are %s"
            % (bad_types, _SUPPORTED_LABEL_TYPES)
        )

    return label_types


def _verify_image_ids(
    selected_split_ids, unspecified_ids, dataset_dir, split, download=True
):
    # Download all image IDs, verify given IDs, sort unspecified IDs into
    # current split

    split_ids, did_download = _load_all_image_ids(
        dataset_dir, split=split, download=download
    )

    # Verify image IDs are in correct split
    sid_set = set(split_ids)
    ssid_set = set(selected_split_ids)
    verified_split_ids = sid_set & ssid_set
    incorrect_split_ids = ssid_set - verified_split_ids
    if incorrect_split_ids:
        logger.warning(
            "Found %d IDs that do not belong to split '%s': %s",
            len(incorrect_split_ids),
            split,
            incorrect_split_ids,
        )

    # Find any unspecified IDs in this split and add them
    uids_set = set(unspecified_ids)
    unspecified_ids_in_split = sid_set & uids_set

    split_image_ids = list(verified_split_ids) + list(unspecified_ids_in_split)

    return split_image_ids, did_download


def _get_downloaded_image_ids(dataset_dir):
    data_dir = os.path.join(dataset_dir, "data")
    if not os.path.exists(data_dir):
        return []

    return [os.path.splitext(n)[0] for n in etau.list_files(data_dir)]


def _get_all_label_data(
    dataset_dir,
    image_ids,
    label_types=None,
    classes=None,
    oi_classes=None,
    attrs=None,
    oi_attrs=None,
    seg_classes=None,
    download_only=False,
    ids_only=False,
    track_all_ids=True,
    only_matching=False,
    split=None,
    download=False,
):
    cls_data = {}
    det_data = {}
    rel_data = {}
    seg_data = {}

    all_classes_ids = set(image_ids)
    any_classes_ids = set()

    all_attrs_ids = set(image_ids)
    any_attrs_ids = set()

    _label_types = _parse_label_types(label_types)

    did_download = False

    if "classifications" in _label_types:
        if download:
            url = _ANNOTATION_DOWNLOAD_URLS[split]["classifications"]
        else:
            url = None

        cls_all_ids, cls_any_ids, cls_data, _did_download = _get_label_data(
            dataset_dir,
            image_ids,
            "classifications",
            classes=classes,
            oi_classes=oi_classes,
            download_only=download_only,
            ids_only=ids_only,
            track_all_ids=track_all_ids,
            only_matching=only_matching,
            url=url,
            download=download,
        )
        did_download |= _did_download

        # Classifications only capture the label schema, so don't use them for
        # ID list purposes unless they were the only label type requested
        if len(_label_types) == 1:
            all_classes_ids &= cls_all_ids
            any_classes_ids |= cls_any_ids

    if "detections" in _label_types:
        if download:
            url = _ANNOTATION_DOWNLOAD_URLS[split]["detections"]
        else:
            url = None

        det_all_ids, det_any_ids, det_data, _did_download = _get_label_data(
            dataset_dir,
            image_ids,
            "detections",
            classes=classes,
            oi_classes=oi_classes,
            url=url,
            download=download,
            download_only=download_only,
            ids_only=ids_only,
            track_all_ids=track_all_ids,
            only_matching=only_matching,
        )
        did_download |= _did_download

        all_classes_ids &= det_all_ids
        any_classes_ids |= det_any_ids

    if "relationships" in _label_types:
        if download:
            url = _ANNOTATION_DOWNLOAD_URLS[split]["relationships"]
        else:
            url = None

        rel_all_ids, rel_any_ids, rel_data, _did_download = _get_label_data(
            dataset_dir,
            image_ids,
            "relationships",
            classes=attrs,
            oi_classes=oi_attrs,
            download_only=download_only,
            ids_only=ids_only,
            track_all_ids=track_all_ids,
            only_matching=only_matching,
            url=url,
            download=download,
        )
        did_download |= _did_download

        all_attrs_ids &= rel_all_ids
        any_attrs_ids |= rel_any_ids

    if "segmentations" in _label_types:
        if classes is not None and label_types is not None:
            non_seg_classes = sorted(set(classes) - set(seg_classes))
            if non_seg_classes:
                logger.warning(
                    "No segmentations exist for classes %s\nYou can view the "
                    "available segmentation classes via "
                    "`get_segmentation_classes()`",
                    non_seg_classes,
                )

        if download:
            url = _ANNOTATION_DOWNLOAD_URLS[split]["segmentations"]["mask_csv"]
        else:
            url = None

        seg_all_ids, seg_any_ids, seg_data, _did_download = _get_label_data(
            dataset_dir,
            image_ids,
            "segmentations",
            classes=classes,
            oi_classes=oi_classes,
            download_only=download_only,
            ids_only=ids_only,
            track_all_ids=track_all_ids,
            only_matching=only_matching,
            url=url,
            download=download,
        )
        did_download |= _did_download

        all_classes_ids &= seg_all_ids
        any_classes_ids |= seg_any_ids

    if classes is not None:
        all_label_ids = all_classes_ids
        any_label_ids = any_classes_ids

        if attrs is not None:
            all_label_ids &= all_attrs_ids
            any_label_ids &= any_attrs_ids
    elif attrs is not None:
        all_label_ids = all_attrs_ids
        any_label_ids = any_attrs_ids
    else:
        all_label_ids = all_classes_ids & all_attrs_ids
        any_label_ids = any_classes_ids | any_attrs_ids

    return (
        cls_data,
        det_data,
        rel_data,
        seg_data,
        all_label_ids,
        any_label_ids,
        did_download,
    )


def _get_label_data(
    dataset_dir,
    image_ids,
    label_type,
    classes=None,
    oi_classes=None,
    download_only=False,
    ids_only=False,
    track_all_ids=True,
    only_matching=False,
    url=None,
    download=True,
):
    csv_path = os.path.join(dataset_dir, "labels", label_type + ".csv")
    did_download = _download_file_if_necessary(
        csv_path, url, quiet=0, download=download
    )

    if download_only:
        return set(), set(), {}, did_download

    df = _parse_csv(csv_path, dataframe=True)
    df.set_index("ImageID", drop=False, inplace=True)
    df = df.loc[df.index.intersection(image_ids)]

    if classes is not None:
        # Restrict by classes
        if label_type == "relationships":
            cols = ["LabelName1", "LabelName2"]
        else:
            cols = ["LabelName"]

        oi_classes = set(oi_classes)

        if track_all_ids:
            observed = defaultdict(set)
            for image_id, labels in zip(df["ImageID"].values, df[cols].values):
                observed[image_id].update(labels)

            all_ids = set()
            any_ids = set()
            for image_id, observed_classes in observed.items():
                if oi_classes.issubset(observed_classes):
                    all_ids.add(image_id)

                if oi_classes & observed_classes:
                    any_ids.add(image_id)
        else:
            any_df = df[df[cols].isin(oi_classes).any(axis=1)]
            all_ids = set()
            any_ids = set(any_df["ImageID"].unique())
    else:
        # No class restriction
        all_ids = set()
        any_ids = set(df["ImageID"].unique())

    if ids_only:
        return all_ids, any_ids, {}, did_download

    if classes is not None:
        if only_matching:
            # Only keep the specified labels
            relevant_df = df[df[cols].isin(oi_classes).any(axis=1)].copy()
        else:
            # Keep all labels for the relevant image IDs
            relevant_df = df.loc[df.index.intersection(any_ids)]
    else:
        relevant_df = df

    relevant_df.sort_index(inplace=True)

    data = {
        "all_ids": set(df["ImageID"].unique()),
        "relevant_ids": any_ids,
        "df": relevant_df,
    }

    return all_ids, any_ids, data, did_download


def _download(
    image_ids,
    downloaded_ids,
    oi_classes,
    oi_attrs,
    seg_classes,
    dataset_dir,
    split,
    label_types=None,
    classes=None,
    attrs=None,
    max_samples=None,
    shuffle=False,
    num_workers=None,
    download=True,
):
    # Download any necessary labels, and, if specific classes/attributes are
    # requested, determine which image IDs have the specified labels
    (
        _,
        _,
        _,
        _,
        all_label_ids,
        any_label_ids,
        did_download,
    ) = _get_all_label_data(
        dataset_dir,
        image_ids,
        label_types=label_types,
        classes=classes,
        oi_classes=oi_classes,
        attrs=attrs,
        oi_attrs=oi_attrs,
        seg_classes=seg_classes,
        download_only=classes is None and attrs is None,
        ids_only=True,
        track_all_ids=max_samples is not None,
        split=split,
        download=download,
    )

    downloaded_ids = set(downloaded_ids)

    # Make list of `target_ids`
    if classes is not None or attrs is not None:
        if max_samples is not None:
            #
            # Bias sampling to meet user requirements per the priorities below:
            # (1) samples with all labels
            # (2) already downloaded samples with relevant labels
            # (3) non-downloaded samples with relevant labels
            #

            not_all_ids = any_label_ids - all_label_ids
            existing_ids = not_all_ids & downloaded_ids
            non_existing_ids = not_all_ids - downloaded_ids

            all_label_ids = sorted(all_label_ids)
            existing_ids = sorted(existing_ids)
            non_existing_ids = sorted(non_existing_ids)

            if shuffle:
                random.shuffle(all_label_ids)
                random.shuffle(existing_ids)
                random.shuffle(non_existing_ids)

            target_ids = all_label_ids + existing_ids + non_existing_ids
            target_ids = target_ids[:max_samples]
        else:
            # Include all samples that meet any requirement
            target_ids = sorted(any_label_ids)

            if shuffle:
                random.shuffle(target_ids)
    else:
        if max_samples is not None:
            # Bias sampling towards already-downloaded samples
            image_ids = set(image_ids)
            existing_ids = sorted(image_ids & downloaded_ids)
            non_existing_ids = sorted(image_ids - downloaded_ids)

            if shuffle:
                random.shuffle(existing_ids)
                random.shuffle(non_existing_ids)

            target_ids = existing_ids + non_existing_ids
            target_ids = target_ids[:max_samples]
        else:
            # Use all available samples
            target_ids = sorted(image_ids)

            if shuffle:
                random.shuffle(target_ids)

    num_target = len(target_ids)
    all_ids = list(downloaded_ids | set(target_ids))
    num_samples = len(all_ids)  # total downloaded

    if max_samples is not None and num_target < max_samples:
        logger.warning(
            "Only found %d (<%d) samples matching your requirements",
            num_target,
            max_samples,
        )

    if "segmentations" in _parse_label_types(label_types):
        _did_download = _download_masks_if_necessary(
            all_ids, dataset_dir, split, download=download
        )
        did_download |= _did_download

    num_downloaded = _download_images_if_necessary(
        target_ids,
        split,
        dataset_dir,
        num_workers=num_workers,
        download=download,
    )

    if num_downloaded > 0:
        did_download = True

    return num_samples, did_download


def _get_dataframe_rows(df, image_id):
    left = df["ImageID"].searchsorted(image_id, "left")
    right = df["ImageID"].searchsorted(image_id, "right")
    return df[left:right]


def _create_classifications(cls_data, image_id, classes_map):
    all_label_ids = cls_data["all_ids"]
    relevant_ids = cls_data["relevant_ids"]
    df = cls_data["df"]

    if image_id not in all_label_ids:
        return None, None

    if image_id not in relevant_ids:
        pos_labels = fol.Classifications()
        neg_labels = fol.Classifications()
        return pos_labels, neg_labels

    def _make_label(row):
        # [ImageID,Source,LabelName,Confidence]
        return fol.Classification(
            label=classes_map[row["LabelName"]],
            confidence=float(row["Confidence"]),
        )

    matching_df = _get_dataframe_rows(df, image_id)
    cls = [_make_label(row[1]) for row in matching_df.iterrows()]

    pos_cls = []
    neg_cls = []
    for c in cls:
        if c.confidence > 0.1:
            pos_cls.append(c)
        else:
            neg_cls.append(c)

    pos_labels = fol.Classifications(classifications=pos_cls)
    neg_labels = fol.Classifications(classifications=neg_cls)

    return pos_labels, neg_labels


def _create_detections(det_data, image_id, classes_map):
    all_label_ids = det_data["all_ids"]
    relevant_ids = det_data["relevant_ids"]
    df = det_data["df"]

    if image_id not in all_label_ids:
        return None

    if image_id not in relevant_ids:
        return fol.Detections()

    def _make_label(row):
        # ImageID,Source,LabelName,Confidence,XMin,XMax,YMin,YMax,IsOccluded,IsTruncated,IsGroupOf,IsDepiction,IsInside

        xmin = float(row["XMin"])
        xmax = float(row["XMax"])
        ymin = float(row["YMin"])
        ymax = float(row["YMax"])
        bounding_box = [xmin, ymin, xmax - xmin, ymax - ymin]

        return fol.Detection(
            bounding_box=bounding_box,
            label=classes_map[row["LabelName"]],
            IsOccluded=bool(int(row["IsOccluded"])),
            IsTruncated=bool(int(row["IsTruncated"])),
            IsGroupOf=bool(int(row["IsGroupOf"])),
            IsDepiction=bool(int(row["IsDepiction"])),
            IsInside=bool(int(row["IsInside"])),
        )

    matching_df = _get_dataframe_rows(df, image_id)
    dets = [_make_label(row[1]) for row in matching_df.iterrows()]
    return fol.Detections(detections=dets)


def _create_relationships(rel_data, image_id, classes_map, attrs_map):
    all_label_ids = rel_data["all_ids"]
    relevant_ids = rel_data["relevant_ids"]
    df = rel_data["df"]

    if image_id not in all_label_ids:
        return None

    if image_id not in relevant_ids:
        return fol.Detections()

    def _make_label(row):
        # ImageID,LabelName1,LabelName2,XMin1,XMax1,YMin1,YMax1,XMin2,XMax2,YMin2,YMax2,RelationshipLabel

        oi_label1 = row["LabelName1"]
        xmin1 = float(row["XMin1"])
        xmax1 = float(row["XMax1"])
        ymin1 = float(row["YMin1"])
        ymax1 = float(row["YMax1"])

        oi_label2 = row["LabelName2"]
        xmin2 = float(row["XMin2"])
        xmax2 = float(row["XMax2"])
        ymin2 = float(row["YMin2"])
        ymax2 = float(row["YMax2"])

        if oi_label1 in classes_map:
            label1 = classes_map[oi_label1]
        else:
            label1 = attrs_map[oi_label1]

        if oi_label2 in classes_map:
            label2 = classes_map[oi_label2]
        else:
            label2 = attrs_map[oi_label2]

        label_rel = row["RelationshipLabel"]

        xmin_int = min(xmin1, xmin2)
        ymin_int = min(ymin1, ymin2)
        xmax_int = max(xmax1, xmax2)
        ymax_int = max(ymax1, ymax2)

        bounding_box = [
            xmin_int,
            ymin_int,
            xmax_int - xmin_int,
            ymax_int - ymin_int,
        ]

        return fol.Detection(
            bounding_box=bounding_box,
            label=label_rel,
            Label1=label1,
            Label2=label2,
        )

    matching_df = _get_dataframe_rows(df, image_id)
    rels = [_make_label(row[1]) for row in matching_df.iterrows()]
    return fol.Detections(detections=rels)


def _create_segmentations(seg_data, image_id, classes_map, dataset_dir):
    all_label_ids = seg_data["all_ids"]
    relevant_ids = seg_data["relevant_ids"]
    df = seg_data["df"]

    if image_id not in all_label_ids:
        return None

    if image_id not in relevant_ids:
        return fol.Detections()

    def _make_label(row):
        # MaskPath,ImageID,LabelName,BoxID,BoxXMin,BoxXMax,BoxYMin,BoxYMax,PredictedIoU,Clicks

        mask_path = row["MaskPath"]
        label = classes_map[row["LabelName"]]
        xmin = float(row["BoxXMin"])
        xmax = float(row["BoxXMax"])
        ymin = float(row["BoxYMin"])
        ymax = float(row["BoxYMax"])

        # Convert to [top-left-x, top-left-y, width, height]
        bbox = [xmin, ymin, xmax - xmin, ymax - ymin]

        # Load boolean mask
        mask_path = os.path.join(
            dataset_dir,
            "labels",
            "masks",
            image_id[0].upper(),
            mask_path,
        )
        if not os.path.isfile(mask_path):
            msg = "Segmentation file %s does not exist", mask_path
            warnings.warn(msg)
            return None

        rgb_mask = etai.read(mask_path)
        mask = etai.rgb_to_gray(rgb_mask) > 122
        h, w = mask.shape
        cropped_mask = mask[
            int(ymin * h) : int(ymax * h), int(xmin * w) : int(xmax * w)
        ]

        return fol.Detection(bounding_box=bbox, label=label, mask=cropped_mask)

    matching_df = _get_dataframe_rows(df, image_id)
    segs = [_make_label(row[1]) for row in matching_df.iterrows()]
    segs = [s for s in segs if s is not None]
    return fol.Detections(detections=segs)


def _load_all_image_ids(dataset_dir, split=None, download=True):
    csv_filepath = os.path.join(dataset_dir, "metadata", "image_ids.csv")
    url = _ANNOTATION_DOWNLOAD_URLS[split]["image_ids"]

    quiet = -1 if split == "train" else 0
    did_download = _download_file_if_necessary(
        csv_filepath, url, quiet=quiet, download=download
    )

    csv_data = _parse_csv(csv_filepath)
    image_ids = [i[0].strip() for i in csv_data[1:]]

    return image_ids, did_download


def _download_file_if_necessary(
    filepath, url, is_zip=False, quiet=-1, download=True
):
    did_download = False

    if is_zip:
        # Check if unzipped directory exists
        unzipped_dir = os.path.splitext(filepath)[0]
        if os.path.isdir(unzipped_dir):
            return did_download

    if not os.path.isfile(filepath):
        if not download:
            raise ValueError("File '%s' is not downloaded" % filepath)

        if quiet < 1:
            logger.info("Downloading '%s' to '%s'", url, filepath)

        etau.ensure_basedir(filepath)
        etaw.download_file(url, path=filepath, quiet=quiet != -1)
        did_download = True

    if is_zip:
        etau.extract_zip(filepath, outdir=unzipped_dir, delete_zip=True)

    return did_download


def _download_masks_if_necessary(image_ids, dataset_dir, split, download=True):
    seg_zip_names = list({i[0].upper() for i in image_ids})
    mask_urls = _ANNOTATION_DOWNLOAD_URLS[split]["segmentations"]["mask_data"]
    masks_dir = os.path.join(dataset_dir, "labels", "masks")

    quiet = 1 if split == "validation" else 0
    did_download = False
    for zip_name in seg_zip_names:
        url = mask_urls[zip_name]
        zip_path = os.path.join(masks_dir, zip_name + ".zip")
        _did_download = _download_file_if_necessary(
            zip_path, url, is_zip=True, quiet=quiet, download=download
        )
        did_download |= _did_download

    return did_download


def _download_images_if_necessary(
    image_ids, split, dataset_dir, num_workers=None, download=True
):
    data_dir = os.path.join(dataset_dir, "data")
    etau.ensure_dir(data_dir)

    urls = {}
    num_existing = 0
    for image_id in image_ids:
        filepath = os.path.join(data_dir, image_id + ".jpg")
        obj = split + "/" + image_id + ".jpg"  # AWS path, always use "/"
        if not os.path.isfile(filepath):
            url = "s3://%s/%s" % (_BUCKET_NAME, obj)
            urls[url] = filepath
        else:
            num_existing += 1

    num_images = len(urls)

    if num_images == 0:
        if download:
            logger.info("Necessary images already downloaded")

        return num_images

    if not download:
        raise ValueError("%d images are not downloaded" % num_images)

    if num_existing > 0:
        logger.info(
            "Found %d images, downloading the remaining %d",
            num_existing,
            num_images,
        )
    else:
        logger.info("Downloading %d images", num_images)

    foua.download_public_s3_files(urls, num_workers=num_workers)

    return num_images


def _verify_version(version):
    if version not in _SUPPORTED_VERSIONS:
        raise ValueError(
            "Version %s is not supported. Supported versions are: %s"
            % (version, ", ".join(_SUPPORTED_VERSIONS))
        )


_ANNOTATION_DOWNLOAD_URLS = {
    "general": {
        "class_names": "https://storage.googleapis.com/openimages/v5/class-descriptions-boxable.csv",
        "attr_names": "https://storage.googleapis.com/openimages/v6/oidv6-attributes-description.csv",
        "hierarchy": "https://storage.googleapis.com/openimages/2018_04/bbox_labels_600_hierarchy.json",
        "segmentation_classes": "https://storage.googleapis.com/openimages/v5/classes-segmentation.txt",
    },
    "train": {
        "num_images": 1743042,
        "image_ids": "https://storage.googleapis.com/openimages/2018_04/train/train-images-boxable-with-rotation.csv",
        "classifications": "https://storage.googleapis.com/openimages/v5/train-annotations-human-imagelabels-boxable.csv",
        "detections": "https://storage.googleapis.com/openimages/v6/oidv6-train-annotations-bbox.csv",
        "relationships": "https://storage.googleapis.com/openimages/v6/oidv6-train-annotations-vrd.csv",
        "segmentations": {
            "mask_csv": "https://storage.googleapis.com/openimages/v5/train-annotations-object-segmentation.csv",
            "mask_data": {
                "0": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-0.zip",
                "1": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-1.zip",
                "2": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-2.zip",
                "3": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-3.zip",
                "4": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-4.zip",
                "5": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-5.zip",
                "6": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-6.zip",
                "7": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-7.zip",
                "8": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-8.zip",
                "9": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-9.zip",
                "A": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-a.zip",
                "B": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-b.zip",
                "C": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-c.zip",
                "D": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-d.zip",
                "E": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-e.zip",
                "F": "https://storage.googleapis.com/openimages/v5/train-masks/train-masks-f.zip",
            },
        },
    },
    "test": {
        "num_images": 125436,
        "image_ids": "https://storage.googleapis.com/openimages/2018_04/test/test-images-with-rotation.csv",
        "classifications": "https://storage.googleapis.com/openimages/v5/test-annotations-human-imagelabels-boxable.csv",
        "detections": "https://storage.googleapis.com/openimages/v5/test-annotations-bbox.csv",
        "relationships": "https://storage.googleapis.com/openimages/v6/oidv6-test-annotations-vrd.csv",
        "segmentations": {
            "mask_csv": "https://storage.googleapis.com/openimages/v5/test-annotations-object-segmentation.csv",
            "mask_data": {
                "0": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-0.zip",
                "1": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-1.zip",
                "2": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-2.zip",
                "3": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-3.zip",
                "4": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-4.zip",
                "5": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-5.zip",
                "6": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-6.zip",
                "7": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-7.zip",
                "8": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-8.zip",
                "9": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-9.zip",
                "A": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-a.zip",
                "B": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-b.zip",
                "C": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-c.zip",
                "D": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-d.zip",
                "E": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-e.zip",
                "F": "https://storage.googleapis.com/openimages/v5/test-masks/test-masks-f.zip",
            },
        },
    },
    "validation": {
        "num_images": 41620,
        "image_ids": "https://storage.googleapis.com/openimages/2018_04/validation/validation-images-with-rotation.csv",
        "classifications": "https://storage.googleapis.com/openimages/v5/validation-annotations-human-imagelabels-boxable.csv",
        "detections": "https://storage.googleapis.com/openimages/v5/validation-annotations-bbox.csv",
        "relationships": "https://storage.googleapis.com/openimages/v6/oidv6-validation-annotations-vrd.csv",
        "segmentations": {
            "mask_csv": "https://storage.googleapis.com/openimages/v5/validation-annotations-object-segmentation.csv",
            "mask_data": {
                "0": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-0.zip",
                "1": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-1.zip",
                "2": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-2.zip",
                "3": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-3.zip",
                "4": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-4.zip",
                "5": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-5.zip",
                "6": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-6.zip",
                "7": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-7.zip",
                "8": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-8.zip",
                "9": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-9.zip",
                "A": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-a.zip",
                "B": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-b.zip",
                "C": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-c.zip",
                "D": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-d.zip",
                "E": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-e.zip",
                "F": "https://storage.googleapis.com/openimages/v5/validation-masks/validation-masks-f.zip",
            },
        },
    },
}

_BUCKET_NAME = "open-images-dataset"

_CSV_DELIMITERS = [",", ";", ":", " ", "\t", "\n"]

_SUPPORTED_LABEL_TYPES = [
    "classifications",
    "detections",
    "relationships",
    "segmentations",
]

_SUPPORTED_SPLITS = [
    "train",
    "test",
    "validation",
]

_SUPPORTED_VERSIONS = [
    "v6",
]



"""
Downloads a subset of CIFAR-100 and stores it to disk as follows::

    /tmp/fiftyone/
    └── cifar100_with_duplicates/
        ├── <classA>/
        │   ├── <image1>.jpg
        │   ├── <image2>.jpg
        │   └── ...
        ├── <classB>/
        │   ├── <image1>.jpg
        │   ├── <image2>.jpg
        │   └── ...
        └── ...

A random 5% of the samples are duplicates, instead of the original samples.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
import random

import eta.core.image as etai
import eta.core.utils as etau

import fiftyone.core.utils as fou

fou.ensure_tf()
from tensorflow.keras.datasets import cifar100


DATASET_SIZE = 1000
DATASET_DIR = os.path.join("/tmp/fiftyone/cifar100_with_duplicates")
CORRUPTION_RATE = 0.05

FINE_CLASSES = [
    "apple",
    "aquarium_fish",
    "baby",
    "bear",
    "beaver",
    "bed",
    "bee",
    "beetle",
    "bicycle",
    "bottle",
    "bowl",
    "boy",
    "bridge",
    "bus",
    "butterfly",
    "camel",
    "can",
    "castle",
    "caterpillar",
    "cattle",
    "chair",
    "chimpanzee",
    "clock",
    "cloud",
    "cockroach",
    "couch",
    "crab",
    "crocodile",
    "cup",
    "dinosaur",
    "dolphin",
    "elephant",
    "flatfish",
    "forest",
    "fox",
    "girl",
    "hamster",
    "house",
    "kangaroo",
    "computer_keyboard",
    "lamp",
    "lawn_mower",
    "leopard",
    "lion",
    "lizard",
    "lobster",
    "man",
    "maple_tree",
    "motorcycle",
    "mountain",
    "mouse",
    "mushroom",
    "oak_tree",
    "orange",
    "orchid",
    "otter",
    "palm_tree",
    "pear",
    "pickup_truck",
    "pine_tree",
    "plain",
    "plate",
    "poppy",
    "porcupine",
    "possum",
    "rabbit",
    "raccoon",
    "ray",
    "road",
    "rocket",
    "rose",
    "sea",
    "seal",
    "shark",
    "shrew",
    "skunk",
    "skyscraper",
    "snail",
    "snake",
    "spider",
    "squirrel",
    "streetcar",
    "sunflower",
    "sweet_pepper",
    "table",
    "tank",
    "telephone",
    "television",
    "tiger",
    "tractor",
    "train",
    "trout",
    "tulip",
    "turtle",
    "wardrobe",
    "whale",
    "willow_tree",
    "wolf",
    "woman",
    "worm",
]


def download_dataset():
    print(
        "Downloading dataset of %d samples to:\n\t%s"
        % (DATASET_SIZE, DATASET_DIR)
    )
    print(
        "and corrupting the data (%d%% duplicates)" % (100 * CORRUPTION_RATE)
    )

    # if not empty, delete current contents
    etau.ensure_empty_dir(DATASET_DIR, cleanup=True)

    (_, _), (x_test, y_test) = cifar100.load_data(label_mode="fine")

    dataset_size = min(DATASET_SIZE, 10000)

    x = x_test[:dataset_size, :]
    y = y_test[:dataset_size, :]

    for i in range(x.shape[0]):
        if random.random() > 0.95:
            # pick a random sample 5% of the time
            idx = random.randint(0, x.shape[0])
        else:
            idx = i

        # get label
        fine_label = FINE_CLASSES[y[idx, 0]]

        # read image
        img = x[idx, :]

        rel_img_path = os.path.join(fine_label, "%d.jpg" % i)
        abs_img_path = os.path.join(DATASET_DIR, rel_img_path)

        etai.write(img, abs_img_path)

    print("Download successful")


if __name__ == "__main__":
    download_dataset()



"""
TensorFlow utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import contextlib
import logging
import multiprocessing
import os
import warnings

import cv2
import numpy as np
from skimage.color import rgba2rgb

import eta.core.image as etai
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
import fiftyone.core.utils as fou
import fiftyone.utils.data as foud

fou.ensure_tf(eager=True)
import tensorflow as tf


logger = logging.getLogger(__name__)


def from_images_dir(
    images_dir, recursive=True, force_rgb=False, num_parallel_calls=None
):
    """Creates a ``tf.data.Dataset`` for the given directory of images.

    Args:
        images_dir: a directory of images
        recursive (True): whether to recursively traverse subdirectories
        force_rgb (False): whether to force convert all images to RGB
        num_parallel_calls (None): the number of samples to read
            asynchronously in parallel. See
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map for
            details

    Returns:
        a ``tf.data.Dataset`` that emits decoded images
    """
    image_paths = foud.parse_images_dir(images_dir, recursive=recursive)
    return from_images(
        image_paths, force_rgb=force_rgb, num_parallel_calls=num_parallel_calls
    )


def from_images_patt(images_patt, force_rgb=False, num_parallel_calls=None):
    """Creates a ``tf.data.Dataset`` for the given glob pattern of images.

    Args:
        images_patt: a glob pattern of images like ``/path/to/images/*.jpg``
        force_rgb (False): whether to force convert all images to RGB
        num_parallel_calls (None): the number of samples to read
            asynchronously in parallel. See
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map for
            details

    Returns:
        a ``tf.data.Dataset`` that emits decoded images
    """
    image_paths = etau.get_glob_matches(images_patt)
    return from_images(
        image_paths, force_rgb=force_rgb, num_parallel_calls=num_parallel_calls
    )


def from_images(image_paths, force_rgb=False, num_parallel_calls=None):
    """Creates a ``tf.data.Dataset`` for the given list of images.

    Args:
        image_paths: an iterable of image paths
        force_rgb (False): whether to force convert all images to RGB
        num_parallel_calls (None): the number of samples to read
            asynchronously in parallel. See
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map for
            details

    Returns:
        a ``tf.data.Dataset`` that emits decoded images
    """

    def parse_sample(image_path):
        return _parse_image_tf(image_path, force_rgb=force_rgb)

    return tf.data.Dataset.from_tensor_slices(list(image_paths)).map(
        parse_sample, num_parallel_calls=num_parallel_calls
    )


def from_image_paths_and_labels(
    image_paths, labels, force_rgb=False, num_parallel_calls=None
):
    """Creates a ``tf.data.Dataset`` for an image classification dataset stored
    as a list of image paths and labels.

    Args:
        image_paths: an iterable of image paths
        labels: an iterable of labels
        force_rgb (False): whether to force convert all images to RGB
        num_parallel_calls (None): the number of samples to read
            asynchronously in parallel. See
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map for
            details

    Returns:
        a ``tf.data.Dataset`` that emits ``(img, label)`` pairs
    """

    def parse_sample(image_path, label):
        img = _parse_image_tf(image_path, force_rgb=force_rgb)
        return img, label

    return tf.data.Dataset.from_tensor_slices(
        (list(image_paths), list(labels))
    ).map(parse_sample, num_parallel_calls=num_parallel_calls)


def from_image_classification_dir_tree(
    dataset_dir, force_rgb=False, num_parallel_calls=None
):
    """Creates a ``tf.data.Dataset`` for the given image classification dataset
    directory tree.

    The directory should have the following format::

        <dataset_dir>/
            <classA>/
                <image1>.<ext>
                <image2>.<ext>
                ...
            <classB>/
                <image1>.<ext>
                <image2>.<ext>
                ...

    Args:
        dataset_dir: the dataset directory
        force_rgb (False): whether to force convert all images to RGB
        num_parallel_calls (None): the number of samples to read
            asynchronously in parallel. See
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset#map for
            details

    Returns:
        a tuple of

        -   **dataset**: a ``tf.data.Dataset` that emits ``(img, label)`` pairs
        -   **classes**: a list of class label strings
    """
    samples, classes = foud.parse_image_classification_dir_tree(dataset_dir)

    def parse_sample(sample):
        image_path, label = sample
        img = _parse_image_tf(image_path, force_rgb=force_rgb)
        return img, label

    dataset = tf.data.Dataset.from_tensor_slices(samples).map(
        parse_sample, num_parallel_calls=num_parallel_calls
    )
    return dataset, classes


def from_tf_records(
    tf_records_patt, buffer_size=None, num_parallel_reads=None
):
    """Creates a ``tf.data.Dataset`` for the TFRecords at the given path(s).

    Args:
        tf_records_patt: the path (or glob pattern of paths) to the TFRecords
            file(s) to load
        buffer_size (None): an optional buffer size, in bytes, to use when
            reading the records. Reasonable values are 1-100MBs
        num_parallel_reads (None): an optional number of files to read in
            parallel. If a negative value is passed, this parameter is set to
            the number of CPU cores on the host machine. By default, the files
            are read in series

    Returns:
        a ``tf.data.Dataset`` that emits ``tf.train.Example`` protos
    """
    if num_parallel_reads is not None and num_parallel_reads < 0:
        num_parallel_reads = multiprocessing.cpu_count()

    return tf.data.TFRecordDataset(
        tf.data.Dataset.list_files(tf_records_patt),
        buffer_size=buffer_size,
        num_parallel_reads=num_parallel_reads,
    )


def write_tf_records(examples, tf_records_path, num_shards=None):
    """Writes the given ``tf.train.Example`` protos to disk as TFRecords.

    Args:
        examples: an iterable that emits ``tf.train.Example`` protos
        tf_records_path: the path to write the ``.tfrecords`` file. If sharding
            is requested ``-%%05d-of-%%05d`` is appended to the path
        num_shards (None): an optional number of shards to split the records
            into (using a round robin strategy)
    """
    with TFRecordsWriter(tf_records_path, num_shards=num_shards) as writer:
        for example in examples:
            writer.write(example)


class TFRecordsWriter(object):
    """Class for writing TFRecords to disk.

    Example Usage::

        with TFRecordsWriter("/path/for/tf.records", num_shards=5) as writer:
            for tf_example in tf_examples:
                writer.write(tf_example)

    Args:
        tf_records_path: the path to write the ``.tfrecords`` file. If sharding
            is requested ``-%%05d-of-%%05d`` is appended to the path
        num_shards (None): an optional number of shards to split the records
            into (using a round robin strategy). If omitted, no sharding is
            used
    """

    def __init__(self, tf_records_path, num_shards=None):
        self.tf_records_path = tf_records_path
        self.num_shards = num_shards

        self._idx = None
        self._num_shards = None
        self._writers = None
        self._writers_context = None

    def __enter__(self):
        self._idx = -1

        etau.ensure_basedir(self.tf_records_path)

        if self.num_shards:
            self._num_shards = self.num_shards
            tf_records_patt = self.tf_records_path + "-%05d-of-%05d"
            tf_records_paths = [
                tf_records_patt % (i, self.num_shards)
                for i in range(self.num_shards)
            ]
        else:
            self._num_shards = 1
            tf_records_paths = [self.tf_records_path]

        self._writers_context = contextlib.ExitStack()
        c = self._writers_context.__enter__()

        self._writers = [
            c.enter_context(tf.io.TFRecordWriter(path))
            for path in tf_records_paths
        ]

        return self

    def __exit__(self, *args):
        self._writers_context.__exit__(*args)

    def write(self, tf_example):
        """Writres the ``tf.train.Example`` proto to disk.

        Args:
            tf_example: a ``tf.train.Example`` proto
        """
        self._idx += 1
        self._writers[self._idx % self._num_shards].write(
            tf_example.SerializeToString()
        )


class TFRecordSampleParser(foud.LabeledImageSampleParser):
    """Base class for sample parsers that ingest ``tf.train.Example`` protos
    containing labeled images.

    Args:
        force_rgb (False): whether to force convert all images to RGB
    """

    # Subclasses must implement this
    _FEATURES = {}

    def __init__(self, force_rgb=False):
        super().__init__()
        self.force_rgb = force_rgb

        self._current_features_cache = None
        self._channels = 3 if force_rgb else 0

    def get_image(self):
        return self._parse_image(self._current_features)

    def get_label(self):
        return self._parse_label(self._current_features)

    def clear_sample(self):
        super().clear_sample()
        self._current_features_cache = None

    @property
    def _current_features(self):
        if self._current_features_cache is None:
            self._current_features_cache = self._parse_features(
                self.current_sample
            )

        return self._current_features_cache

    def _parse_features(self, sample):
        return tf.io.parse_single_example(sample, self._FEATURES)

    def _parse_image(self, features):
        raise NotImplementedError("subclasses must implement _parse_image()")

    def _parse_label(self, features):
        raise NotImplementedError("subclasses must implement _parse_label()")


class TFImageClassificationSampleParser(TFRecordSampleParser):
    """Parser for image classification samples stored as
    `TFRecords <https://www.tensorflow.org/tutorials/load_data/tfrecord>`_.

    This implementation supports samples that are ``tf.train.Example`` protos
    whose features follow the format described in
    :ref:`this page <TFImageClassificationDataset-import>`.

    Args:
        force_rgb (False): whether to force convert all images to RGB
    """

    _FEATURES = {
        "height": tf.io.FixedLenFeature([], tf.int64),
        "width": tf.io.FixedLenFeature([], tf.int64),
        "depth": tf.io.FixedLenFeature([], tf.int64),
        "filename": tf.io.FixedLenFeature([], tf.string),
        "format": tf.io.FixedLenFeature([], tf.string),
        "image_bytes": tf.io.FixedLenFeature([], tf.string),
        "label": tf.io.FixedLenFeature([], tf.string, default_value=""),
    }

    @property
    def label_cls(self):
        return fol.Classification

    @property
    def has_image_path(self):
        return False

    @property
    def has_image_metadata(self):
        return True

    def get_image_metadata(self):
        return self._parse_image_metadata(self._current_features)

    def _parse_image(self, features):
        img_bytes = features["image_bytes"]
        img = tf.image.decode_image(img_bytes, channels=self._channels)
        return img.numpy()

    def _parse_image_metadata(self, features):
        if self.force_rgb:
            num_channels = 3
        else:
            num_channels = features["depth"].numpy()

        return fom.ImageMetadata(
            size_bytes=len(features["image_bytes"].numpy()),
            mime_type="image/" + features["format"].numpy().decode(),
            width=features["width"].numpy(),
            height=features["height"].numpy(),
            num_channels=num_channels,
        )

    def _parse_label(self, features):
        label = features["label"].numpy().decode()
        if not label:
            return None

        return fol.Classification(label=label)


class TFObjectDetectionSampleParser(TFRecordSampleParser):
    """Parser for samples in
    `TF Object Detection API format <https://github.com/tensorflow/models/blob/master/research/object_detection>`_.

    This implementation supports samples that are ``tf.train.Example`` protos
    whose features follow the format described in
    :ref:`this page <TFObjectDetectionDataset-import>`.

    Args:
        force_rgb (False): whether to force convert all images to RGB
    """

    _FEATURES = {
        "image/height": tf.io.FixedLenFeature([], tf.int64),
        "image/width": tf.io.FixedLenFeature([], tf.int64),
        "image/filename": tf.io.FixedLenFeature([], tf.string),
        "image/source_id": tf.io.FixedLenFeature([], tf.string),
        "image/encoded": tf.io.FixedLenFeature([], tf.string),
        "image/format": tf.io.FixedLenFeature([], tf.string),
        "image/object/bbox/xmin": tf.io.FixedLenSequenceFeature(
            [],
            tf.float32,
            allow_missing=True,
        ),
        "image/object/bbox/xmax": tf.io.FixedLenSequenceFeature(
            [], tf.float32, allow_missing=True
        ),
        "image/object/bbox/ymin": tf.io.FixedLenSequenceFeature(
            [], tf.float32, allow_missing=True
        ),
        "image/object/bbox/ymax": tf.io.FixedLenSequenceFeature(
            [], tf.float32, allow_missing=True
        ),
        "image/object/class/text": tf.io.FixedLenSequenceFeature(
            [], tf.string, allow_missing=True
        ),
        "image/object/class/label": tf.io.FixedLenSequenceFeature(
            [], tf.int64, allow_missing=True
        ),
    }

    @property
    def label_cls(self):
        return fol.Detections

    @property
    def has_image_path(self):
        return False

    @property
    def has_image_metadata(self):
        return True

    def get_image_metadata(self):
        return self._parse_image_metadata(self._current_features)

    def _parse_image(self, features):
        img_bytes = features["image/encoded"]
        img = tf.image.decode_image(img_bytes, channels=self._channels)
        return img.numpy()

    def _parse_image_metadata(self, features):
        if self.force_rgb:
            num_channels = 3
        else:
            num_channels = None

        return fom.ImageMetadata(
            size_bytes=len(features["image/encoded"].numpy()),
            mime_type="image/" + features["image/format"].numpy().decode(),
            width=features["image/width"].numpy(),
            height=features["image/height"].numpy(),
            num_channels=num_channels,
        )

    def _parse_label(self, features):
        xmins = features["image/object/bbox/xmin"].numpy()
        xmaxs = features["image/object/bbox/xmax"].numpy()
        ymins = features["image/object/bbox/ymin"].numpy()
        ymaxs = features["image/object/bbox/ymax"].numpy()
        texts = features["image/object/class/text"].numpy()
        detections = []
        for xmin, xmax, ymin, ymax, text in zip(
            xmins, xmaxs, ymins, ymaxs, texts
        ):
            label = text.decode()
            detections.append(
                fol.Detection(
                    label=label,
                    bounding_box=[
                        float(xmin),
                        float(ymin),
                        float(xmax - xmin),
                        float(ymax - ymin),
                    ],
                )
            )

        return fol.Detections(detections=detections)


class TFRecordsLabeledImageDatasetImporter(
    foud.LabeledImageDatasetImporter, foud.ImportPathsMixin
):
    """Base class for
    :class:`fiftyone.utils.data.importers.LabeledImageDatasetImporter`
    instances that import ``tf.train.Example`` protos containing labeled
    images.

    This class assumes that the input TFRecords only contain the images
    themselves and not their paths on disk, and, therefore, the images are read
    in-memory and written to the provided ``images_dir`` during import.

    Args:
        dataset_dir (None): the dataset directory. If omitted,
            ``tf_records_path`` must be provided
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` or glob pattern like
                ``"*.records-*-of-*"`` specifying the location of the records
                in ``dataset_dir``
            -   an absolute filepath or glob pattern for the records. In this
                case, ``dataset_dir`` has no effect on the location of the
                records

            If None, the parameter will default to ``*record*``
        images_dir (None): the directory in which the images will be written.
            If not provided, the images will be unpacked into ``dataset_dir``
        image_format (None): the image format to use to write the images to
            disk. By default, ``fiftyone.config.default_image_ext`` is used
        force_rgb (False): whether to force convert all images to RGB
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        tf_records_path=None,
        images_dir=None,
        image_format=None,
        force_rgb=False,
        max_samples=None,
    ):
        if dataset_dir is None and tf_records_path is None:
            raise ValueError(
                "Either `dataset_dir` or `tf_records_path` must be provided"
            )

        tf_records_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=tf_records_path,
            default="*record*",
        )

        if images_dir is None:
            images_dir = os.path.dirname(tf_records_path)
            logger.warning(
                "No `images_dir` provided. Images will be unpacked to '%s'",
                images_dir,
            )

        super().__init__(dataset_dir=dataset_dir, max_samples=max_samples)

        self.tf_records_path = tf_records_path
        self.images_dir = images_dir
        self.image_format = image_format
        self.force_rgb = force_rgb

        self._sample_parser = self._make_sample_parser()
        self._dataset_ingestor = None
        self._iter_dataset_ingestor = None

    def __iter__(self):
        self._iter_dataset_ingestor = iter(self._dataset_ingestor)
        return self

    def __next__(self):
        return next(self._iter_dataset_ingestor)

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_image_metadata(self):
        return self._sample_parser.has_image_metadata

    def setup(self):
        self._dataset_ingestor = foud.LabeledImageDatasetIngestor(
            self.images_dir,
            from_tf_records(self.tf_records_path),
            self._sample_parser,
            image_format=self.image_format,
            max_samples=self.max_samples,
        )
        self._dataset_ingestor.setup()

    def close(self, *args):
        self._dataset_ingestor.close(*args)

    def _make_sample_parser(self):
        """Returns a :class:`TFRecordSampleParser` instance for parsing
        TFRecords read by this importer.
        """
        raise NotImplementedError(
            "subclasses must implement _make_sample_parser()"
        )


class TFImageClassificationDatasetImporter(
    TFRecordsLabeledImageDatasetImporter
):
    """Importer for TF image classification datasets stored on disk.

    This class assumes that the input TFRecords only contain the images
    themselves and not their paths on disk, and, therefore, the images are read
    in-memory and written to the provided ``images_dir`` during import.

    See :ref:`this page <TFImageClassificationDataset-import>` for format
    details.

    Args:
        dataset_dir (None): the dataset directory. If omitted,
            ``tf_records_path`` must be provided
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` or glob pattern like
                ``"*.records-*-of-*"`` specifying the location of the records
                in ``dataset_dir``
            -   an absolute filepath or glob pattern for the records. In this
                case, ``dataset_dir`` has no effect on the location of the
                records

            If None, the parameter will default to ``*record*``
        images_dir (None): the directory in which the images will be written.
            If not provided, the images will be unpacked into ``dataset_dir``
        image_format (None): the image format to use to write the images to
            disk. By default, ``fiftyone.config.default_image_ext`` is used
        force_rgb (False): whether to force convert all images to RGB
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    @property
    def label_cls(self):
        return fol.Classification

    def _make_sample_parser(self):
        return TFImageClassificationSampleParser(force_rgb=self.force_rgb)


class TFObjectDetectionDatasetImporter(TFRecordsLabeledImageDatasetImporter):
    """Importer for TF detection datasets stored on disk.

    This class assumes that the input TFRecords only contain the images
    themselves and not their paths on disk, and, therefore, the images are read
    in-memory and written to the provided ``images_dir`` during import.

    See :ref:`this page <TFObjectDetectionDataset-import>` for format
    details.

    Args:
        dataset_dir (None): the dataset directory. If omitted,
            ``tf_records_path`` must be provided
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` or glob pattern like
                ``"*.records-*-of-*"`` specifying the location of the records
                in ``dataset_dir``
            -   an absolute filepath or glob pattern for the records. In this
                case, ``dataset_dir`` has no effect on the location of the
                records

            If None, the parameter will default to ``*record*``
        images_dir (None): the directory in which the images will be written.
            If not provided, the images will be unpacked into ``dataset_dir``
        image_format (None): the image format to use to write the images to
            disk. By default, ``fiftyone.config.default_image_ext`` is used
        force_rgb (False): whether to force convert all images to RGB
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    @property
    def label_cls(self):
        return fol.Detections

    def _make_sample_parser(self):
        return TFObjectDetectionSampleParser(force_rgb=self.force_rgb)


class TFRecordsDatasetExporter(
    foud.LabeledImageDatasetExporter, foud.ExportPathsMixin
):
    """Base class for
    :class:`fiftyone.utils.data.exporters.LabeledImageDatasetExporter`
    instances that export labeled images as TFRecords datasets on disk.

    Args:
        export_dir (None): the directory to write the export. This has no
            effect if ``tf_records_path`` is an absolute path
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` specifying the location of
                the records in ``export_dir``
            -   an absolute filepath for the records. In this case,
                ``export_dir`` has no effect on the location of the records

            If None, the parameter will default to ``tf.records``
        num_shards (None): an optional number of shards to split the records
            into (using a round robin strategy). If specified,
            ``-%%05d-of-%%05d`` is appended to the records path
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        force_rgb (False): whether to force convert all images to RGB
    """

    def __init__(
        self,
        export_dir=None,
        tf_records_path=None,
        num_shards=None,
        image_format=None,
        force_rgb=False,
    ):
        tf_records_path = self._parse_labels_path(
            export_dir=export_dir,
            labels_path=tf_records_path,
            default="tf.records",
        )

        if image_format is None:
            image_format = fo.config.default_image_ext

        super().__init__(export_dir=export_dir)

        self.tf_records_path = tf_records_path
        self.num_shards = num_shards
        self.image_format = image_format
        self.force_rgb = force_rgb

        self._example_generator = None
        self._filename_maker = None
        self._tf_records_writer = None

    @property
    def requires_image_metadata(self):
        return False

    def setup(self):
        self._example_generator = self._make_example_generator()
        self._filename_maker = fou.UniqueFilenameMaker(
            default_ext=self.image_format
        )
        self._tf_records_writer = TFRecordsWriter(
            self.tf_records_path, num_shards=self.num_shards
        )
        self._tf_records_writer.__enter__()

    def export_sample(self, image_or_path, label, metadata=None):
        if etau.is_str(image_or_path):
            filename = image_or_path
        else:
            filename = self._filename_maker.get_output_path()

        tf_example = self._example_generator.make_tf_example(
            image_or_path, label, filename=filename
        )
        self._tf_records_writer.write(tf_example)

    def close(self, *args):
        self._tf_records_writer.__exit__(*args)

    def _make_example_generator(self):
        """Returns a :class:`TFExampleGenerator` instance that will generate
        ``tf.train.Example`` protos for this exporter.
        """
        raise NotImplementedError(
            "subclasses must implement _make_example_generator()"
        )


class TFImageClassificationDatasetExporter(TFRecordsDatasetExporter):
    """Exporter that writes an image classification dataset to disk as
    TFRecords.

    See :ref:`this page <TFImageClassificationDataset-export>` for format
    details.

    Args:
        export_dir (None): the directory to write the export. Can be omitted if
            ``tf_records_path`` is provided
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` specifying the location of
                the records in ``export_dir``
            -   an absolute filepath for the records. In this case,
                ``export_dir`` has no effect on the location of the records

            If None, the parameter will default to ``tf.records``
        num_shards (None): an optional number of shards to split the records
            into (using a round robin strategy). If specified,
            ``-%%05d-of-%%05d`` is appended to the records path
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        force_rgb (False): whether to force convert all images to RGB
    """

    @property
    def label_cls(self):
        return fol.Classification

    def _make_example_generator(self):
        return TFImageClassificationExampleGenerator(force_rgb=self.force_rgb)


class TFObjectDetectionDatasetExporter(TFRecordsDatasetExporter):
    """Exporter that writes an object detection dataset to disk as TFRecords
    in the TF Object Detection API format.

    See :ref:`this page <TFObjectDetectionDataset-export>` for format details.

    Args:
        export_dir (None): the directory to write the export. Can be omitted if
            ``tf_records_path`` is provided
        tf_records_path (None): an optional parameter that enables explicit
            control over the location of the TF records. Can be any of the
            following:

            -   a filename like ``"tf.records"`` specifying the location of
                the records in ``export_dir``
            -   an absolute filepath for the records. In this case,
                ``export_dir`` has no effect on the location of the records

            If None, the parameter will default to ``tf.records``
        num_shards (None): an optional number of shards to split the records
            into (using a round robin strategy). If specified,
            ``-%%05d-of-%%05d`` is appended to the records path
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        force_rgb (False): whether to force convert all images to RGB
        classes (None): the list of possible class labels
    """

    def __init__(
        self,
        export_dir=None,
        tf_records_path=None,
        num_shards=None,
        image_format=None,
        force_rgb=False,
        classes=None,
    ):
        super().__init__(
            export_dir=export_dir,
            tf_records_path=tf_records_path,
            num_shards=num_shards,
            image_format=image_format,
            force_rgb=force_rgb,
        )

        self.classes = classes

    @property
    def label_cls(self):
        return fol.Detections

    def _make_example_generator(self):
        return TFObjectDetectionExampleGenerator(
            force_rgb=self.force_rgb, classes=self.classes
        )


class TFExampleGenerator(object):
    """Base class for sample writers that emit ``tf.train.Example`` protos.

    Args:
        force_rgb (False): whether to force convert all images to RGB
    """

    def __init__(self, force_rgb=False):
        self.force_rgb = force_rgb

    def make_tf_example(self, image_or_path, label, *args, **kwargs):
        """Makes a ``tf.train.Example`` for the given data.

        Args:
            image_or_path: an image or the path to the image on disk
            label: a :class:`fiftyone.core.labels.Label`
            *args: subclass-specific positional arguments
            **kwargs: subclass-specific keyword arguments

        Returns:
            a ``tf.train.Example``
        """
        raise NotImplementedError(
            "subclasses must implement make_tf_example()"
        )

    def _parse_image_or_path(self, image_or_path, filename=None):
        if etau.is_str(image_or_path):
            image_path = image_or_path

            if filename is None:
                filename = os.path.basename(image_path)

            # pylint: disable=no-member
            img_bytes = etau.read_file(image_path, binary=True)
            img = etai.decode(img_bytes, flag=cv2.IMREAD_ANYCOLOR)

            if img.ndim == 2:
                img = np.expand_dims(img, 2)

            if self.force_rgb and img.shape[2] == 1:
                img = img.repeat(3, axis=2)
                img_bytes = etai.encode(img, os.path.splitext(image_path)[1])
        else:
            img = image_or_path

            if filename is None:
                raise ValueError(
                    "`filename` must be provided when `image_or_path` is an "
                    "image"
                )

            if img.ndim == 2:
                img = np.expand_dims(img, 2)

            if self.force_rgb and img.shape[2] == 1:
                img = img.repeat(3, axis=2)
            elif self.force_rgb and img.shape[2] == 4:
                img = rgba2rgb(img)

            if filename.endswith((".jpg", ".jpeg")):
                img_bytes = tf.image.encode_jpeg(img)
            elif filename.endswith(".png"):
                img_bytes = tf.image.encode_png(img)
            else:
                raise ValueError(
                    "Unsupported image format '%s'"
                    % os.path.splitext(filename)[1]
                )

        img_shape = img.shape

        return img_bytes, img_shape, filename


class TFImageClassificationExampleGenerator(TFExampleGenerator):
    """Class for generating ``tf.train.Example`` protos for samples in TF
    image classification format.

    See :ref:`this page <TFImageClassificationDataset-export>` for format
    details.

    Args:
        force_rgb (False): whether to force convert all images to RGB
    """

    def make_tf_example(self, image_or_path, classification, filename=None):
        """Makes a ``tf.train.Example`` for the given data.

        Args:
            image_or_path: an image or the path to the image on disk
            classification: a :class:`fiftyone.core.labels.Classification`
                instance, or ``None``
            filename (None): a filename for the image. Required when
                ``image_or_path`` is an image, in which case the extension of
                this filename determines the encoding used. If
                ``image_or_path`` is the path to an image, this is optional; by
                default, the basename of ``image_path`` is used

        Returns:
            a ``tf.train.Example``
        """
        img_bytes, img_shape, filename = self._parse_image_or_path(
            image_or_path, filename=filename
        )
        format = os.path.splitext(filename)[1][1:]  # no leading `.`

        feature = {
            "height": _int64_feature(img_shape[0]),
            "width": _int64_feature(img_shape[1]),
            "depth": _int64_feature(img_shape[2]),
            "filename": _bytes_feature(filename.encode()),
            "format": _bytes_feature(format.encode()),
            "image_bytes": _bytes_feature(img_bytes),
        }

        if classification is not None:
            label = classification.label
            feature["label"] = _bytes_feature(label.encode())

        return tf.train.Example(features=tf.train.Features(feature=feature))


class TFObjectDetectionExampleGenerator(TFExampleGenerator):
    """Class for generating ``tf.train.Example`` protos for samples in TF
    Object Detection API format.

    See :ref:`this page <TFObjectDetectionDataset-export>` for format details.

    Args:
        force_rgb (False): whether to force convert all images to RGB
        classes (None): the list of possible class labels
    """

    def __init__(self, force_rgb=False, classes=None):
        super().__init__(force_rgb=force_rgb)

        self.classes = classes

        if classes:
            dynamic_classes = False
            labels_map_rev = _to_labels_map_rev(classes)
        else:
            dynamic_classes = True
            labels_map_rev = {}

        self._dynamic_classes = dynamic_classes
        self._labels_map_rev = labels_map_rev

    def make_tf_example(self, image_or_path, detections, filename=None):
        """Makes a ``tf.train.Example`` for the given data.

        Args:
            image_or_path: an image or the path to the image on disk
            detections: a :class:`fiftyone.core.labels.Detections` instance, or
                ``None``
            filename (None): a filename for the image. Required when
                ``image_or_path`` is an image, in which case the extension of
                this filename determines the encoding used. If
                ``image_or_path`` is the path to an image, this is optional; by
                default, the basename of ``image_path`` is used

        Returns:
            a ``tf.train.Example``
        """
        img_bytes, img_shape, filename = self._parse_image_or_path(
            image_or_path, filename=filename
        )
        format = os.path.splitext(filename)[1][1:]  # no leading `.`

        feature = {
            "image/height": _int64_feature(img_shape[0]),
            "image/width": _int64_feature(img_shape[1]),
            "image/filename": _bytes_feature(filename.encode()),
            "image/source_id": _bytes_feature(filename.encode()),
            "image/encoded": _bytes_feature(img_bytes),
            "image/format": _bytes_feature(format.encode()),
        }

        if detections is not None:
            xmins, xmaxs, ymins, ymaxs, texts, labels = [], [], [], [], [], []
            for detection in detections.detections:
                xmin, ymin, w, h = detection.bounding_box
                text = detection.label

                if self._dynamic_classes:
                    if text not in self._labels_map_rev:
                        label = len(self._labels_map_rev)
                        self._labels_map_rev[text] = label
                    else:
                        label = self._labels_map_rev[text]
                elif text not in self._labels_map_rev:
                    msg = (
                        "Ignoring detection with label '%s' not in provided "
                        "classes" % text
                    )
                    warnings.warn(msg)
                    continue
                else:
                    label = self._labels_map_rev[text]

                xmins.append(xmin)
                xmaxs.append(xmin + w)
                ymins.append(ymin)
                ymaxs.append(ymin + h)
                texts.append(text.encode())
                labels.append(label)

            feature.update(
                {
                    "image/object/bbox/xmin": _float_list_feature(xmins),
                    "image/object/bbox/xmax": _float_list_feature(xmaxs),
                    "image/object/bbox/ymin": _float_list_feature(ymins),
                    "image/object/bbox/ymax": _float_list_feature(ymaxs),
                    "image/object/class/text": _bytes_list_feature(texts),
                    "image/object/class/label": _int64_list_feature(labels),
                }
            )

        return tf.train.Example(features=tf.train.Features(feature=feature))


def _get_classes_for_detections(samples, label_field):
    classes = set()
    for sample in samples:
        for detection in sample[label_field].detections:
            classes.add(detection.label)

    return sorted(classes)


def _to_labels_map_rev(classes):
    return {c: i for i, c in enumerate(classes)}


def _parse_image_tf(image_path, force_rgb=False):
    channels = 3 if force_rgb else 0
    img_bytes = tf.io.read_file(image_path)
    return tf.image.decode_image(img_bytes, channels=channels)


def _bytes_feature(value):
    # Extract bytes from EagerTensor, if necessary
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()

    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _bytes_list_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def _float_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _float_list_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _int64_list_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))



"""
FiftyOne Server /sort route

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request

import fiftyone.core.dataset as fod
import fiftyone.core.fields as fof
import fiftyone.core.view as fov

from fiftyone.server.decorators import route
import fiftyone.server.events as fose
from fiftyone.server.query import serialize_dataset
import fiftyone.server.view as fosv


class Sort(HTTPEndpoint):
    @route
    async def post(self, request: Request, data: dict):
        dataset_name = data.get("dataset", None)
        filters = data.get("filters", {})
        stages = data.get("view", None)
        extended = data.get("extended", None)
        dist_field = data.get("dist_field", None)

        dataset = fod.load_dataset(dataset_name)

        changed = False
        if dist_field and not dataset.get_field(dist_field):
            dataset.add_sample_field(dist_field, fof.FloatField)
            changed = True

        fosv.get_view(dataset_name, stages, filters, extended_stages=None)

        state = fose.get_state().copy()
        view = fosv.get_view(dataset_name, stages, filters)
        state.dataset = view._dataset

        if isinstance(view, fov.DatasetView):
            state.view = view
        else:
            view = None

        return {
            "dataset": await serialize_dataset(dataset_name, stages)
            if changed
            else None,
            "state": state.serialize(),
        }



"""
Session class for interacting with the FiftyOne App.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from functools import wraps
import logging
from packaging.version import Version
import pkg_resources
import time
import typing as t
import webbrowser
from uuid import uuid4

try:
    import IPython.display
except:
    pass

import fiftyone as fo
import fiftyone.constants as focn
import fiftyone.core.dataset as fod
from fiftyone.core.config import AppConfig
import fiftyone.core.context as focx
import fiftyone.core.plots as fop
import fiftyone.core.service as fos
import fiftyone.core.utils as fou
import fiftyone.core.view as fov
from fiftyone.core.state import StateDescription

import fiftyone.core.session.client as fosc
from fiftyone.core.session.events import (
    CaptureNotebookCell,
    CloseSession,
    DeactivateNotebookCell,
    ReactivateNotebookCell,
    StateUpdate,
)
import fiftyone.core.session.notebooks as fosn


logger = logging.getLogger(__name__)

#
# Session globals
#
# _session is the proxy `Session` for `launch_app()` and `close_app()` calls
# _server_services maintains active servers
# _subscribed_sessions maintains sessions subscribed to an active server
#
# Both maps use port as the key, so the main python process is always aware
# of what servers can be killed
#
# Note that a server process is killed via deletion of its
# `fiftyone.core.service.ServerService` instance
#
_session = None
_server_services = {}
_subscribed_sessions = defaultdict(set)

_APP_DESKTOP_MESSAGE = """
Desktop App launched.
"""

_APP_WEB_MESSAGE = """
App launched. Point your web browser to http://localhost:{0}
"""

_APP_NOTEBOOK_MESSAGE = """
Session launched. Run `session.show()` to open the App in a cell output.
"""

_REMOTE_INSTRUCTIONS = """
You have launched a remote App on port {0}. To connect to this App from another
machine, issue the following command to configure port forwarding:

ssh -N -L 5151:127.0.0.1:{0} [<username>@]<hostname>

where `[<username>@]<hostname>` refers to your current machine. The App can
then be viewed in your browser at http://localhost:5151.

Alternatively, if you have FiftyOne installed on your local machine, just run:

fiftyone app connect --destination [<username>@]<hostname> --port {0}

See https://voxel51.com/docs/fiftyone/user_guide/app.html#remote-sessions
for more information about remote sessions.
"""

_WAIT_INSTRUCTIONS = """
A session appears to have terminated shortly after it was started. If you
intended to start an App instance or a remote session from a script, you should
call `session.wait()` to keep the session (and the script) alive.
"""


def launch_app(
    dataset: fod.Dataset = None,
    view: fov.DatasetView = None,
    port: int = None,
    address: str = None,
    remote: bool = False,
    desktop: bool = None,
    height: int = None,
    auto: bool = True,
    config: AppConfig = None,
) -> "Session":
    """Launches the FiftyOne App.

    Note that only one App instance can be opened at a time. If this method is
    called when another App exists, the existing App will be closed.

    Args:
        dataset (None): an optional :class:`fiftyone.core.dataset.Dataset` or
            :class:`fiftyone.core.view.DatasetView` to load
        view (None): an optional :class:`fiftyone.core.view.DatasetView` to
            load
        port (None): the port number to serve the App. If None,
            ``fiftyone.config.default_app_port`` is used
        address (None): the address to serve the App. If None,
            ``fiftyone.config.default_app_address`` is used
        remote (False): whether this is a remote session, and opening the App
            should not be attempted
        desktop (None): whether to launch the App in the browser (False) or as
            a desktop App (True). If None, ``fiftyone.config.desktop_app`` is
            used. Not applicable to notebook contexts
        height (None): an optional height, in pixels, at which to render App
            instances in notebook cells. Only applicable in notebook contexts
        auto (True): whether to automatically show a new App window
            whenever the state of the session is updated. Only applicable
            in notebook contexts
        config (None): an optional :class:`fiftyone.core.config.AppConfig` to
            control fine-grained default App settings

    Returns:
        a :class:`Session`
    """
    global _session  # pylint: disable=global-statement

    #
    # Note, we always `close_app()` here rather than just calling
    # `session.open()` if a session already exists, because the app may have
    # been closed in some way other than `session.close()` --- e.g., the user
    # closing the GUI --- in which case the underlying Electron process may
    # still exist; in this case, `session.open()` does not seem to reopen the
    # app
    #
    # @todo this can probably be improved
    #
    close_app()

    _session = Session(
        dataset=dataset,
        view=view,
        port=port,
        address=address,
        remote=remote,
        desktop=desktop,
        height=height,
        auto=auto,
        config=config,
    )

    if _session.remote:
        logger.info(_REMOTE_INSTRUCTIONS.strip().format(_session.server_port))
    elif _session.desktop:
        logger.info(_APP_DESKTOP_MESSAGE.strip())
    elif focx.is_notebook_context():
        if not auto:
            logger.info(_APP_NOTEBOOK_MESSAGE.strip())
    else:
        logger.info(_APP_WEB_MESSAGE.strip().format(_session.server_port))

    return _session


def close_app() -> None:
    """Closes the FiftyOne App, if necessary.

    If no App is currently open, this method has no effect.
    """
    global _session  # pylint: disable=global-statement
    if _session is not None:
        _session.close()
        _session = None


def update_state(auto_show: bool = False) -> t.Callable:
    """:class:`Session` method decorator for triggering state update events

    Args:
        auto_show (False): whether the method should show a new notebook App
            cell as well, if ``auto`` is ``True``

    Returns:
        the decorated method
    """

    def decorator(func: t.Callable) -> t.Callable:
        @wraps(func)
        def wrapper(
            session: "Session", *args: t.Tuple, **kwargs: dict
        ) -> t.Any:
            if auto_show and session.auto and focx.is_notebook_context():
                session.freeze()
            result = func(session, *args, **kwargs)
            session._client.send_event(StateUpdate(state=session._state))
            if auto_show and session.auto and focx.is_notebook_context():
                session.show()

            return result

        return wrapper

    return decorator


class Session(object):
    """Session that maintains a 1-1 shared state with the FiftyOne App.

    **Basic Usage**

    -   Use :func:`launch_app` to launch the App and retrieve its
        corresponding :class:`Session` instance.

    -   To open a dataset in the App, simply set the
        :attr:`Session.dataset` property of the session to your
        :class:`fiftyone.core.dataset.Dataset`.

    -   To load a specific view into your dataset, simply set the
        :attr:`Session.view` property of the session to your
        :class:`fiftyone.core.view.DatasetView`.

    -   To attach/remove interactive plots, use the methods exposed on the
        :attr:`Session.plots` property of the session.

    -   Use :meth:`Session.refresh` to refresh the App if you update a dataset
        outside of the App

    -   Use :attr:`Session.selected` to retrieve the IDs of the currently
        selected samples in the App.

    -   Use :attr:`Session.selected_labels` to retrieve the IDs of the
        currently selected labels in the App.

    -   In notebook contexts, use :func:`Session.freeze` to replace the App and
        any attached plots with static images.

    -   Use :func:`Session.close` and :func:`Session.open` to temporarily close
        and reopen the App without creating a new :class:`Session`
        instance.

    -   Use :func:`close_app` to programmatically close the App and
        terminate the session.

    Args:
        dataset (None): an optional :class:`fiftyone.core.dataset.Dataset` or
            :class:`fiftyone.core.view.DatasetView` to load
        view (None): an optional :class:`fiftyone.core.view.DatasetView` to
            load
        plots (None): an optional
            :class:`fiftyone.core.plots.manager.PlotManager` to connect to this
            session
        port (None): the port number to serve the App. If None,
            ``fiftyone.config.default_app_port`` is used
        address (None): the address to serve the App. If None,
            ``fiftyone.config.default_app_address`` is used
        remote (False): whether this is a remote session, and opening the App
            should not be attempted
        desktop (None): whether to launch the App in the browser (False) or as
            a desktop App (True). If None, ``fiftyone.config.desktop_app`` is
            used. Not applicable to notebook contexts (e.g., Jupyter and Colab)
        height (None): an optional height, in pixels, at which to render App
            instances in notebook cells. Only applicable in notebook contexts
        auto (True): whether to automatically show a new App window
            whenever the state of the session is updated. Only applicable
            in notebook contexts
        config (None): an optional :class:`fiftyone.core.config.AppConfig` to
            control fine-grained default App settings
    """

    def __init__(
        self,
        dataset: t.Union[fod.Dataset, fov.DatasetView] = None,
        view: fov.DatasetView = None,
        plots: fop.PlotManager = None,
        port: int = None,
        address: str = None,
        remote: bool = False,
        desktop: bool = None,
        height: int = None,
        auto: bool = True,
        config: AppConfig = None,
    ) -> None:
        # Allow `dataset` to be a view
        if isinstance(dataset, fov.DatasetView):
            view = dataset
            dataset = dataset._root_dataset

        self._validate(dataset, view, plots, config)

        if port is None:
            port = fo.config.default_app_port

        if address is None:
            if fou.is_docker():
                address = "0.0.0.0"
            else:
                address = fo.config.default_app_address

        if config is None:
            config = fo.app_config.copy()

        if height is not None:
            config.notebook_height = height

        self._plots: t.Optional[fop.PlotManager] = None
        self._wait_closed = False

        # Maintain a reference to prevent garbage collection
        self._get_time = time.perf_counter
        self._disable_wait_warning = False
        self._notebook_cells: t.Dict[str, fosn.NotebookCell] = {}

        if desktop is None:
            desktop = (
                fo.config.desktop_app
                if not focx.is_notebook_context()
                else False
            )

        self.plots = plots

        self._state = StateDescription(
            config=config,
            dataset=view._root_dataset if view is not None else dataset,
            view=view,
        )
        self._client = fosc.Client(
            address=address,
            auto=auto,
            desktop=desktop,
            port=port,
            remote=remote,
            start_time=self._get_time(),
        )
        self._client.run(self._state)
        _attach_listeners(self)
        _register_session(self)

        if self.auto and focx.is_notebook_context():
            self.show(height=config.notebook_height)

        if self.remote:
            if focx.is_notebook_context():
                raise ValueError(
                    "Remote sessions cannot be run from a notebook"
                )

            return

        if self.desktop:
            if focx.is_notebook_context():
                raise ValueError(
                    "Cannot open a Desktop App instance from a %s notebook"
                    % focx._get_context()
                )

            if not focn.DEV_INSTALL:
                import_desktop()

            self._app_service = fos.AppService(
                server_port=port, server_address=address
            )
            return

        if not focx.is_notebook_context():
            self.open()
            return

    def _validate(
        self,
        dataset: t.Optional[t.Union[fod.Dataset, fov.DatasetView]],
        view: t.Optional[fov.DatasetView],
        plots: t.Optional[fop.PlotManager],
        config: t.Optional[AppConfig],
    ) -> None:
        if dataset is not None and not isinstance(dataset, fod.Dataset):
            raise ValueError(
                "`dataset` must be a %s or None; found %s"
                % (fod.Dataset, type(dataset))
            )

        if view is not None and not isinstance(view, fov.DatasetView):
            raise ValueError(
                "`view` must be a %s or None; found %s"
                % (fov.DatasetView, type(view))
            )

        if plots is not None and not isinstance(plots, fop.PlotManager):
            raise ValueError(
                "`plots` must be a %s or None; found %s"
                % (fop.PlotManager, type(plots))
            )

        if config is not None and not isinstance(config, AppConfig):
            raise ValueError(
                "`config` must be a %s or None; found %s"
                % (AppConfig, type(config))
            )

    def __repr__(self) -> str:
        return self.summary()

    def __del__(self) -> None:
        try:
            if (
                not self._disable_wait_warning
                and self._get_time() - self._client.start_time < 2.5
            ):
                # logger may already have been garbage-collected
                print(_WAIT_INSTRUCTIONS)

            _unregister_session(self)
        except:
            # e.g. globals were already garbage-collected
            pass

    @property
    def auto(self) -> bool:
        """The auto setting for the session."""
        return self._client.auto

    @auto.setter
    def auto(self, auto: bool) -> None:
        self._client.auto = auto

    @property
    def server_port(self) -> int:
        """The server port for the session."""
        return self._client.port

    @property
    def server_address(self) -> str:
        """The server address for the session, or None if not specified."""
        return self._client.address

    @property
    def remote(self) -> bool:
        """Whether the session is remote."""
        return self._client.remote

    @property
    def desktop(self) -> bool:
        """Whether the session is connected to a desktop App."""
        return self._client.desktop

    @property
    def url(self) -> str:
        """The URL of the session."""
        return focx.get_url(self.server_address, self.server_port)

    @property
    def config(self) -> AppConfig:
        """The current :class:`fiftyone.core.config.AppConfig`.

        For changes to a session's config to take effect in the App,
        a call to :meth:`Session.refresh` or another state-updating action
        such as `session.view = my_view` must occur.

        Example usage::

            import fiftyone as fo

            dataset, session = fo.quickstart()

            # change the show confidence setting and push the change to the App
            session.config.show_confidence = False
            session.refresh()
        """
        return self._state.config

    @config.setter  # type: ignore
    @update_state()
    def config(self, config: t.Optional[AppConfig]) -> None:
        if config is None:
            config = fo.app_config.copy()

        if not isinstance(config, AppConfig):
            raise ValueError(
                "`Session.config` must be a %s or None; found %s"
                % (AppConfig, type(config))
            )

        self._state.config = config

    @property
    def _collection(self) -> t.Union[fod.Dataset, fov.DatasetView, None]:
        if self.view is not None:
            return self.view

        return self.dataset

    @property
    def dataset(self) -> t.Union[fod.Dataset, None]:
        """The :class:`fiftyone.core.dataset.Dataset` connected to the session."""
        return self._state.dataset

    @dataset.setter  # type: ignore
    @update_state(auto_show=True)
    def dataset(self, dataset: t.Union[fod.Dataset, None]) -> None:
        if dataset is not None and not isinstance(dataset, fod.Dataset):
            raise ValueError(
                "`Session.dataset` must be a %s or None; found %s"
                % (fod.Dataset, type(dataset))
            )

        if dataset is not None:
            dataset._reload()

        self._state.dataset = dataset
        self._state.view = None
        self._state.selected = []
        self._state.selected_labels = []

    @update_state()
    def clear_dataset(self) -> None:
        """Clears the current :class:`fiftyone.core.dataset.Dataset` from the
        session, if any.
        """
        self._state.dataset = None

    @property
    def view(self) -> t.Union[fov.DatasetView, None]:
        """The :class:`fiftyone.core.view.DatasetView` connected to the
        session, or ``None`` if no view is connected.
        """
        return self._state.view

    @view.setter  # type: ignore
    @update_state(auto_show=True)
    def view(self, view: t.Union[fov.DatasetView, None]) -> None:
        if view is not None and not isinstance(view, fov.DatasetView):
            raise ValueError(
                "`Session.view` must be a %s or None; found %s"
                % (fov.DatasetView, type(view))
            )

        self._state.view = view

        if view is not None:
            view._root_dataset._reload()
            self._state.dataset = view._root_dataset

        self._state.selected = []
        self._state.selected_labels = []

    @update_state()
    def clear_view(self) -> None:
        """Clears the current :class:`fiftyone.core.view.DatasetView` from the
        session, if any.
        """
        self._state.view = None

    @property
    def has_plots(self) -> bool:
        """Whether this session has any attached plots."""
        return bool(self._plots)

    @property
    def plots(self) -> t.Union[fop.PlotManager, None]:
        """The :class:`fiftyone.core.plots.manager.PlotManager` instance that
        manages plots attached to this session.
        """
        return self._plots

    @plots.setter
    def plots(self, plots: t.Optional[fop.PlotManager]) -> None:
        if plots is not None and not isinstance(plots, fop.PlotManager):
            raise ValueError(
                "`Session.plots` must be a %s or None; found %s"
                % (fop.PlotManager, type(plots))
            )

        if plots is None:
            plots = fop.PlotManager(self)
        else:
            plots._set_session(self)

        self._plots = plots

    def refresh(self) -> None:
        """Refreshes the current App window."""
        self._client.send_event(StateUpdate(state=self._state, refresh=True))

    @property
    def selected(self) -> t.List[str]:
        """A list of sample IDs of the currently selected samples in the App,
        if any.
        """
        return list(self._state.selected)

    @selected.setter  # type: ignore
    @update_state()
    def selected(self, sample_ids: t.List[str]) -> None:
        self._state.selected = list(sample_ids) if sample_ids else []

    @update_state()
    def clear_selected(self) -> None:
        """Clears the currently selected samples, if any."""
        self._state.selected = []

    @update_state()
    def select_samples(
        self,
        ids: t.Optional[t.Union[str, t.Iterable[str]]] = None,
        tags: t.Optional[t.Union[str, t.Iterable[str]]] = None,
    ) -> None:
        """Selects the specified samples in the current view in the App,

        Args:
            ids (None): an ID or iterable of IDs of samples to select
            tags (None): a tag or iterable of tags of samples to select
        """
        if tags is not None and self._collection:
            ids = self._collection.match_tags(tags).values("id")

        if ids is None:
            ids = []

        self._state.selected = list(ids)

    @property
    def selected_labels(self) -> t.List[dict]:
        """A list of labels currently selected in the App.

        Items are dictionaries with the following keys:

        -   ``label_id``: the ID of the label
        -   ``sample_id``: the ID of the sample containing the label
        -   ``field``: the field name containing the label
        -   ``frame_number``: the frame number containing the label (only
            applicable to video samples)
        """
        return list(self._state.selected_labels)

    @selected_labels.setter  # type: ignore
    @update_state()
    def selected_labels(self, labels: dict) -> None:
        self._state.selected_labels = list(labels) if labels else []

    @update_state()
    def select_labels(
        self,
        labels: t.Optional[t.List[dict]] = None,
        ids: t.Optional[t.Union[str, t.Iterable[str]]] = None,
        tags: t.Optional[t.Union[str, t.Iterable[str]]] = None,
        fields: t.Optional[t.Union[str, t.Iterable[str]]] = None,
    ) -> None:
        """Selects the specified labels in the current view in the App.

        This method uses the same interface as
        :meth:`fiftyone.core.collections.SampleCollection.select_labels` to
        specify the labels to select.

        Args:
            labels (None): a list of dicts specifying the labels to select
            ids (None): an ID or iterable of IDs of the labels to select
            tags (None): a tag or iterable of tags of labels to select
            fields (None): a field or iterable of fields from which to select
        """
        if labels is None and self._collection:
            labels = self._collection._get_selected_labels(
                ids=ids, tags=tags, fields=fields
            )

        self._state.selected_labels = list(labels or [])

    @update_state()
    def clear_selected_labels(self) -> None:
        """Clears the currently selected labels, if any."""
        self._state.selected_labels = []

    @update_state()
    def tag_selected_samples(self, tag: str) -> None:
        """Adds the tag to the currently selected samples, if necessary.

        The currently selected labels are :attr:`Session.selected`.

        Args:
            tag: a tag
        """
        if self._collection is not None:
            self._collection.select(self.selected).tag_samples(tag)

    @update_state()
    def untag_selected_samples(self, tag: str) -> None:
        """Removes the tag from the currently selected samples, if necessary.

        The currently selected labels are :attr:`Session.selected`.

        Args:
            tag: a tag
        """
        if self._collection is not None:
            self._collection.select(self.selected).untag_samples(tag)

    @update_state()
    def tag_selected_labels(self, tag: str) -> None:
        """Adds the tag to the currently selected labels, if necessary.

        The currently selected labels are :attr:`Session.selected_labels`.

        Args:
            tag: a tag
        """
        if self._collection is not None:
            self._collection.select_labels(
                labels=self.selected_labels
            ).tag_labels(tag)

    @update_state()
    def untag_selected_labels(self, tag: str) -> None:
        """Removes the tag from the currently selected labels, if necessary.

        The currently selected labels are :attr:`Session.selected_labels`.

        Args:
            tag: a tag
        """
        if self._collection is not None:
            self._collection.select_labels(
                labels=self.selected_labels
            ).untag_labels(tag)

    @property
    def selected_view(self) -> t.Optional[fov.DatasetView]:
        """A :class:`fiftyone.core.view.DatasetView` containing the currently
        selected content in the App.

        The selected view is defined as follows:

        -   If both samples and labels are selected, the view will contain only
            the :attr:`selected_labels` from within the :attr:`selected`
            samples
        -   If samples are selected, the view will only contain the
            :attr:`selected` samples
        -   If labels are selected, the view will only contain the
            :attr:`selected_labels`
        -   If no samples or labels are selected, the view will be ``None``
        """
        if self._collection is None:
            return None

        if self.selected:
            view = self._collection.select(self.selected)

            if self.selected_labels:
                return view.select_labels(labels=self.selected_labels)

            return view

        if self.selected_labels:
            return self._collection.select_labels(labels=self.selected_labels)

        return None

    def summary(self) -> str:
        """Returns a string summary of the session.

        Returns:
            a string summary
        """
        if self._collection and self.dataset is not None:
            etype = self._collection._elements_str
            elements = [
                ("Dataset:", self.dataset.name),
                ("Media type:", self.dataset.media_type),
                ("Num %s:" % etype, len(self._collection)),
                ("Selected %s:" % etype, len(self.selected)),
                ("Selected labels:", len(self.selected_labels)),
            ]
        else:
            elements = [("Dataset:", "-")]

        if self.remote:
            type_ = "remote"
        elif focx.is_colab_context():
            type_ = "colab"
        elif focx.is_databricks_context():
            type_ = "databricks"
        elif self.desktop:
            type_ = "desktop"
        else:
            type_ = None

        if type_ is None:
            elements.append(("Session URL:", self.url))
        else:
            elements.append(("Session type:", type_))

        elements = fou.justify_headings(elements)
        lines = ["%s %s" % tuple(e) for e in elements]

        if self.view:
            lines.extend(["View stages:", self.view._make_view_stages_str()])

        if self.plots:
            lines.append(self.plots.summary())

        return "\n".join(lines)

    def open(self) -> None:
        """Opens the App, if necessary.

        The behavior of this method depends on your context:

        -   Notebooks: calls :meth:`Session.show` to open a new App window in
            the output of your current cell
        -   Desktop: opens the desktop App, if necessary
        -   Other (non-remote): opens the App in a new browser tab
        """
        if self.remote:
            logger.warning("Remote sessions cannot open new App windows")
            return

        if self.plots:
            self.plots.connect()

        if focx.is_notebook_context():
            self.show()
            return

        if self.desktop:
            self._app_service.start()
            return

        self.open_tab()

    def open_tab(self) -> None:
        """Opens the App in a new tab of your browser.

        This method can be called from Jupyter notebooks and in desktop App
        mode to override the default location of the App.
        """
        if self.remote:
            logger.warning("Remote sessions cannot open new App windows")
            return

        if focx.is_notebook_context():
            IPython.display.display(
                IPython.display.Javascript(
                    "window.open('{url}');".format(url=self.url)
                )
            )
            return

        webbrowser.open(self.url, new=2)

    @update_state()
    def show(self, height: int = None) -> None:
        """Opens the App in the output of the current notebook cell.

        This method has no effect in non-notebook contexts.

        Args:
            height (None): a height, in pixels, for the App
        """
        if not focx.is_notebook_context() or self.desktop:
            return

        self.freeze()
        if self.dataset is not None:
            self.dataset._reload()

        if height is None:
            height = self.config.notebook_height

        uuid = str(uuid4())
        self._notebook_cells[uuid] = fosn.NotebookCell(
            address=self.server_address,
            handle=IPython.display.DisplayHandle(display_id=uuid),
            height=height,
            port=self.server_port,
            subscription=uuid,
        )

        fosn.display(self._client, self._notebook_cells[uuid])

    def no_show(self) -> fou.SetAttributes:
        """Returns a context manager that temporarily prevents new App
        instances from being opened in the current notebook cell when methods
        are run that normally would show new App windows.

        This method has no effect in non-notebook contexts.

        Examples::

            import fiftyone as fo

            dataset = foz.load_zoo_dataset("quickstart")
            session = fo.launch_app(dataset)

            # (new cell)

            # Opens a new App instance
            session.view = dataset.take(100)

            # (new cell)

            # Does not open a new App instance
            with session.no_show():
                session.view = dataset.take(100)

        Returns:
            a context manager
        """
        return fou.SetAttributes(self, auto=False)

    def wait(self, wait: float = 3) -> None:
        """Blocks execution until the App is closed by the user.

        For browser Apps, all connected windows (tabs) must be closed before
        this method will unblock.

        For desktop Apps, all positive ``wait`` values are equivalent;
        execution will immediately unblock when the App is closed.

        Args:
            wait (3): the number of seconds to wait for a new App connection
                before returning if all connections are lost. If negative, the
                process will wait forever, regardless of connections
        """
        if focx.is_notebook_context():
            logger.warning("Notebook sessions cannot wait")
            return

        try:
            if wait < 0:
                while True:
                    time.sleep(10)
            elif self.remote or not self._client.desktop:
                self._wait_closed = False
                while not self._wait_closed:
                    time.sleep(wait)
            else:
                self._app_service.wait()
        except KeyboardInterrupt:
            self._disable_wait_warning = True
            raise

    def close(self) -> None:
        """Closes the session and terminates the App, if necessary."""
        if self.remote:
            return

        if self._client._connected and focx._get_context() == focx._NONE:
            self._client.send_event(CloseSession())

        self.plots.disconnect()

    def freeze(self) -> None:
        """Screenshots the active App cell, replacing it with a static image.

        Only applicable to notebook contexts.
        """
        if not focx.is_notebook_context():
            logger.warning("Only notebook sessions can be frozen")
            return

        self._client.send_event(DeactivateNotebookCell())
        self.plots.freeze()


def _attach_listeners(session: "Session"):
    on_close_session: t.Callable[[CloseSession], None] = lambda event: setattr(
        session, "_wait_closed", True
    )
    session._client.add_event_listener("close_session", on_close_session)

    on_state_update: t.Callable[[StateUpdate], None] = lambda event: setattr(
        session, "_state", event.state
    )
    session._client.add_event_listener("state_update", on_state_update)

    if focx.is_notebook_context() and not focx.is_colab_context():

        def on_capture_notebook_cell(event: CaptureNotebookCell) -> None:
            event.subscription in session._notebook_cells and fosn.capture(
                session._notebook_cells[event.subscription], event
            )

        session._client.add_event_listener(
            "capture_notebook_cell", on_capture_notebook_cell
        )

        def on_reactivate_notebook_cell(event: ReactivateNotebookCell) -> None:
            fosn.display(
                session._client,
                session._notebook_cells[event.subscription],
                reactivate=True,
            )

        session._client.add_event_listener(
            "reactivate_notebook_cell", on_reactivate_notebook_cell
        )


def import_desktop() -> None:
    """Attempts to import :mod:`fiftyone.desktop`

    Raises:
        RuntimeError: If matching ``fiftyone-desktop`` version is not
        installed
    """
    try:
        # pylint: disable=unused-import
        import fiftyone.desktop
    except ImportError as e:
        raise RuntimeError(
            "You must `pip install fiftyone[desktop]` in order to launch the "
            "desktop App"
        ) from e

    # Get `fiftyone-desktop` requirement for current `fiftyone` install
    fiftyone_dist = pkg_resources.get_distribution("fiftyone")
    requirements = fiftyone_dist.requires(extras=["desktop"])
    desktop_req = [r for r in requirements if r.name == "fiftyone-desktop"][0]

    desktop_dist = pkg_resources.get_distribution("fiftyone-desktop")

    if not desktop_req.specifier.contains(
        Version(desktop_dist.version).base_version
    ):
        raise RuntimeError(
            "fiftyone==%s requires fiftyone-desktop%s, but you have "
            "fiftyone-desktop==%s installed.\n"
            "Run `pip install fiftyone[desktop]` to install the proper "
            "desktop package version"
            % (
                fiftyone_dist.version,
                desktop_req.specifier,
                desktop_dist.version,
            )
        )


def _register_session(session: Session) -> None:
    global _server_services  # pylint: disable=global-statement
    if session.server_port not in _server_services:
        _server_services[session.server_port] = fos.ServerService(
            session.server_port,
            address=session.server_address,
            do_not_track=fo.config.do_not_track,
        )

    global _subscribed_sessions  # pylint: disable=global-statement
    _subscribed_sessions[session.server_port].add(session)


def _unregister_session(session: Session) -> None:
    global _subscribed_sessions  # pylint: disable=global-statement
    _subscribed_sessions[session.server_port].discard(session)

    if len(_subscribed_sessions[session.server_port]) == 0:
        global _server_services  # pylint: disable=global-statement
        if session.server_port in _server_services:
            service = _server_services.pop(session.server_port)
            service.stop()



"""
FiftyOne JSON handling

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from bson import ObjectId, json_util
from datetime import date, datetime
from json import JSONEncoder
import math
import numpy as np

import eta.core.serial as etas

from fiftyone.core.sample import Sample, SampleView
from fiftyone.core.stages import ViewStage
import fiftyone.core.utils as fou


_MASK_CLASSES = {"Detection", "Heatmap", "Segmentation"}


def _handle_bytes(o):
    for k, v in o.items():
        if isinstance(v, bytes):
            o[k] = str(fou.deserialize_numpy_array(v).shape)
        elif isinstance(v, dict):
            o[k] = _handle_bytes(v)

    return o


def _handle_numpy_array(raw, _cls=None):
    if _cls not in _MASK_CLASSES:
        return str(fou.deserialize_numpy_array(raw).shape)

    array = fou.deserialize_numpy_array(raw)

    if np.isfortran(array):
        array = np.ascontiguousarray(array)

    return fou.serialize_numpy_array(array, ascii=True)


def _handle_date(dt):
    return {
        "_cls": "DateTime",
        "datetime": fou.datetime_to_timestamp(dt),
    }


def _is_invalid_number(value):
    if not isinstance(value, float):
        return False

    return math.isnan(value) or math.isinf(value)


def stringify(d, _cls=None):
    """Converts unsafe JSON types to strings

    Args:
        d: serializable data

    Returns:
        a stringified version of the data
    """
    if isinstance(d, dict):
        return {k: stringify(v, d.get("_cls", None)) for k, v in d.items()}

    if isinstance(d, (list, tuple)):
        return [stringify(v) for v in d]

    if isinstance(d, bytes):
        return _handle_numpy_array(d, _cls)
    elif isinstance(d, (date, datetime)):
        return _handle_date(d)
    elif isinstance(d, ObjectId):
        return str(d)
    elif _is_invalid_number(d):
        return str(d)

    return d


class FiftyOneJSONEncoder(JSONEncoder):
    """JSON encoder for FiftyOne network comms.

    Any classes with non-standard serialization methods should
    be accounted for in the `default()` method.
    """

    def default(self, o):  # pylint: disable=E0202
        """Returns the serialized representation of the objects

        Args:
            o: the object

        Returns:
            str
        """
        if isinstance(o, (Sample, SampleView)):
            return _handle_bytes(o.to_mongo_dict(include_id=True))
        if issubclass(type(o), ViewStage):
            return o._serialize()
        if isinstance(o, ObjectId):
            return str(o)
        if isinstance(o, float):
            return json_util.dumps(o)
        if isinstance(o, etas.Serializable):
            return o.serialize()

        try:
            return super().default(o)
        except:
            return None

    @staticmethod
    def dumps(data, *args, **kwargs) -> str:
        kwargs["cls"] = FiftyOneJSONEncoder
        return json_util.dumps(
            json_util.loads(
                json_util.dumps(data, *args, **kwargs),
                parse_constant=lambda c: c,
            ),
            **kwargs
        )

    @staticmethod
    def loads(*args, **kwargs) -> dict:
        return json_util.loads(*args, **kwargs)

    @staticmethod
    def process(data, *args, **kwargs):
        kwargs["cls"] = FiftyOneJSONEncoder
        return json_util.loads(
            json_util.dumps(data, *args, **kwargs), parse_constant=lambda c: c
        )



"""
FiftyOne Server app.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from datetime import date, datetime
import os
import pathlib

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    RequestResponseEndpoint,
)
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles
from starlette.types import Scope
import strawberry as gql

import fiftyone as fo
import fiftyone.constants as foc
from fiftyone.server.context import GraphQL
from fiftyone.server.extensions import EndSession
from fiftyone.server.mutation import Mutation
from fiftyone.server.query import Query
from fiftyone.server.routes import routes
from fiftyone.server.scalars import Date, DateTime


class Static(StaticFiles):
    async def get_response(self, path: str, scope: Scope) -> Response:
        response = await super().get_response(path, scope)

        if response.status_code == 404:
            path = pathlib.Path(
                *pathlib.Path(path).parts[2:]
            )  # strip dataset/{name}
            response = await super().get_response(path, scope)
            if response.status_code == 404:
                full_path, stat_result = self.lookup_path("index.html")
                return self.file_response(full_path, stat_result, scope)

        return response


class HeadersMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        response = await call_next(request)
        response.headers["x-colab-notebook-cache-control"] = "no-cache"
        return response


schema = gql.Schema(
    mutation=Mutation,
    query=Query,
    extensions=[EndSession],
    scalar_overrides={
        date: Date,
        datetime: DateTime,
    },
)


app = Starlette(
    middleware=[
        Middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["GET", "POST", "HEAD", "OPTIONS"],
            allow_headers=[
                "access-control-allow-origin",
                "authorization",
                "content-type",
            ],
        ),
        Middleware(HeadersMiddleware),
    ],
    debug=foc.DEV_INSTALL,
    routes=[Route(route, endpoint) for route, endpoint in routes]
    + [
        Route(
            "/graphql",
            GraphQL(
                schema,
                graphiql=foc.DEV_INSTALL,
            ),
        ),
        Mount(
            "/plugins",
            app=Static(
                directory=fo.config.plugins_dir,
                html=True,
            ),
            name="plugins",
        ),
        Mount(
            "/",
            app=Static(
                directory=os.path.join(os.path.dirname(__file__), "static"),
                html=True,
            ),
            name="static",
        ),
    ],
)



"""
Image utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import multiprocessing
import os

import eta.core.image as etai
import eta.core.utils as etau

import fiftyone.core.utils as fou
import fiftyone.core.validation as fov


logger = logging.getLogger(__name__)


def reencode_images(
    sample_collection,
    ext=".png",
    force_reencode=True,
    media_field="filepath",
    output_field=None,
    output_dir=None,
    rel_dir=None,
    delete_originals=False,
    num_workers=None,
    skip_failures=False,
):
    """Re-encodes the images in the sample collection to the given format.

    .. note::

        This method will not update the ``metadata`` field of the collection
        after transforming. You can repopulate the ``metadata`` field if needed
        by calling::

            sample_collection.compute_metadata(overwrite=True)

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        ext (".png"): the image format to use (e.g., ".png" or ".jpg")
        force_reencode (True): whether to re-encode images whose extension
            already matches ``ext``
        media_field ("filepath"): the input field containing the image paths to
            transform
        output_field (None): an optional field in which to store the paths to
            the transformed images. By default, ``media_field`` is updated
            in-place
        output_dir (None): an optional output directory in which to write the
            transformed images. If none is provided, the images are updated
            in-place
        rel_dir (None): an optional relative directory to strip from each input
            filepath to generate a unique identifier that is joined with
            ``output_dir`` to generate an output path for each image. This
            argument allows for populating nested subdirectories in
            ``output_dir`` that match the shape of the input paths
        delete_originals (False): whether to delete the original images after
            re-encoding. This parameter has no effect if the images are being
            updated in-place
        num_workers (None): the number of worker processes to use. By default,
            ``multiprocessing.cpu_count()`` is used
        skip_failures (False): whether to gracefully continue without raising
            an error if an image cannot be re-encoded
    """
    fov.validate_image_collection(sample_collection)

    _transform_images(
        sample_collection,
        ext=ext,
        force_reencode=force_reencode,
        media_field=media_field,
        output_field=output_field,
        output_dir=output_dir,
        rel_dir=rel_dir,
        delete_originals=delete_originals,
        num_workers=num_workers,
        skip_failures=skip_failures,
    )


def transform_images(
    sample_collection,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
    ext=None,
    force_reencode=False,
    media_field="filepath",
    output_field=None,
    output_dir=None,
    rel_dir=None,
    delete_originals=False,
    num_workers=None,
    skip_failures=False,
):
    """Transforms the images in the sample collection according to the provided
    parameters.

    .. note::

        This method will not update the ``metadata`` field of the collection
        after transforming. You can repopulate the ``metadata`` field if needed
        by calling::

            sample_collection.compute_metadata(overwrite=True)

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        size (None): an optional ``(width, height)`` for each image. One
            dimension can be -1, in which case the aspect ratio is preserved
        min_size (None): an optional minimum ``(width, height)`` for each
            image. A dimension can be -1 if no constraint should be applied.
            The images are resized (aspect-preserving) if necessary to meet
            this constraint
        max_size (None): an optional maximum ``(width, height)`` for each
            image. A dimension can be -1 if no constraint should be applied.
            The images are resized (aspect-preserving) if necessary to meet
            this constraint
        interpolation (None): an optional ``interpolation`` argument for
            ``cv2.resize()``
        ext (None): an optional image format to re-encode the source images
            into (e.g., ".png" or ".jpg")
        force_reencode (False): whether to re-encode images whose parameters
            already match the specified values
        media_field ("filepath"): the input field containing the image paths to
            transform
        output_field (None): an optional field in which to store the paths to
            the transformed images. By default, ``media_field`` is updated
            in-place
        output_dir (None): an optional output directory in which to write the
            transformed images. If none is provided, the images are updated
            in-place
        rel_dir (None): an optional relative directory to strip from each input
            filepath to generate a unique identifier that is joined with
            ``output_dir`` to generate an output path for each image. This
            argument allows for populating nested subdirectories in
            ``output_dir`` that match the shape of the input paths
        delete_originals (False): whether to delete the original images if any
            transformation was applied. This parameter has no effect if the
            images are being updated in-place
        num_workers (None): the number of worker processes to use. By default,
            ``multiprocessing.cpu_count()`` is used
        skip_failures (False): whether to gracefully continue without raising
            an error if an image cannot be transformed
    """
    fov.validate_image_collection(sample_collection)

    _transform_images(
        sample_collection,
        size=size,
        min_size=min_size,
        max_size=max_size,
        interpolation=interpolation,
        ext=ext,
        force_reencode=force_reencode,
        media_field=media_field,
        output_field=output_field,
        output_dir=output_dir,
        rel_dir=rel_dir,
        delete_originals=delete_originals,
        num_workers=num_workers,
        skip_failures=skip_failures,
    )


def reencode_image(input_path, output_path):
    """Re-encodes the image to the format specified by the given output path.

    Args:
        input_path: the path to the input image
        output_path: the path to write the output image
    """
    _transform_image(input_path, output_path, force_reencode=True)


def transform_image(
    input_path,
    output_path,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
):
    """Transforms the image according to the provided parameters.

    Args:
        input_path: the path to the input image
        output_path: the path to write the output image
        size (None): an optional ``(width, height)`` for the image. One
            dimension can be -1, in which case the aspect ratio is preserved
        min_size (None): an optional minimum ``(width, height)`` for the image.
            A dimension can be -1 if no constraint should be applied. The image
            is resized (aspect-preserving) if necessary to meet this constraint
        max_size (None): an optional maximum ``(width, height)`` for the image.
            A dimension can be -1 if no constraint should be applied. The image
            is resized (aspect-preserving) if necessary to meet this constraint
        interpolation (None): an optional ``interpolation`` argument for
            ``cv2.resize()``
    """
    _transform_image(
        input_path,
        output_path,
        size=size,
        min_size=min_size,
        max_size=max_size,
        interpolation=interpolation,
    )


def _transform_images(
    sample_collection,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
    ext=None,
    force_reencode=False,
    media_field="filepath",
    output_field=None,
    output_dir=None,
    rel_dir=None,
    delete_originals=False,
    num_workers=None,
    skip_failures=False,
):
    ext = _parse_ext(ext)

    if num_workers is None:
        num_workers = multiprocessing.cpu_count()

    if num_workers <= 1:
        _transform_images_single(
            sample_collection,
            size=size,
            min_size=min_size,
            max_size=max_size,
            interpolation=interpolation,
            ext=ext,
            force_reencode=force_reencode,
            media_field=media_field,
            output_field=output_field,
            output_dir=output_dir,
            rel_dir=rel_dir,
            delete_originals=delete_originals,
            skip_failures=skip_failures,
        )
    else:
        _transform_images_multi(
            sample_collection,
            num_workers,
            size=size,
            min_size=min_size,
            max_size=max_size,
            interpolation=interpolation,
            ext=ext,
            force_reencode=force_reencode,
            media_field=media_field,
            output_field=output_field,
            output_dir=output_dir,
            rel_dir=rel_dir,
            delete_originals=delete_originals,
            skip_failures=skip_failures,
        )


def _transform_images_single(
    sample_collection,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
    ext=None,
    force_reencode=False,
    media_field="filepath",
    output_field=None,
    output_dir=None,
    rel_dir=None,
    delete_originals=False,
    skip_failures=False,
):
    if output_field is None:
        output_field = media_field

    diff_field = output_field != media_field

    view = sample_collection.select_fields(media_field)

    with fou.ProgressBar() as pb:
        for sample in pb(view):
            inpath = sample[media_field]

            outpath = _get_outpath(
                inpath, output_dir=output_dir, rel_dir=rel_dir
            )

            if ext is not None:
                outpath = os.path.splitext(outpath)[0] + ext

            _transform_image(
                inpath,
                outpath,
                size=size,
                min_size=min_size,
                max_size=max_size,
                interpolation=interpolation,
                force_reencode=force_reencode,
                delete_original=delete_originals,
                skip_failures=skip_failures,
            )

            if diff_field or outpath != inpath:
                sample[output_field] = outpath
                sample.save()


def _transform_images_multi(
    sample_collection,
    num_workers,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
    ext=None,
    force_reencode=False,
    media_field="filepath",
    output_field=None,
    output_dir=None,
    rel_dir=None,
    delete_originals=False,
    skip_failures=False,
):
    if output_field is None:
        output_field = media_field

    diff_field = output_field != media_field

    sample_ids, inpaths = sample_collection.values(["id", media_field])

    inputs = []
    for sample_id, inpath in zip(sample_ids, inpaths):
        outpath = _get_outpath(inpath, output_dir=output_dir, rel_dir=rel_dir)

        if ext is not None:
            outpath = os.path.splitext(outpath)[0] + ext

        inputs.append(
            (
                sample_id,
                inpath,
                outpath,
                size,
                min_size,
                max_size,
                interpolation,
                force_reencode,
                delete_originals,
                skip_failures,
            )
        )

    outpaths = {}

    try:
        with fou.ProgressBar(inputs) as pb:
            with fou.get_multiprocessing_context().Pool(
                processes=num_workers
            ) as pool:
                for sample_id, inpath, outpath, _ in pb(
                    pool.imap_unordered(_do_transform, inputs)
                ):
                    if diff_field or outpath != inpath:
                        outpaths[sample_id] = outpath
    finally:
        sample_collection.set_values(output_field, outpaths, key_field="id")


def _do_transform(args):
    sample_id, inpath, outpath = args[:3]
    did_transform = _transform_image(inpath, outpath, *args[3:])
    return sample_id, inpath, outpath, did_transform


def _transform_image(
    inpath,
    outpath,
    size=None,
    min_size=None,
    max_size=None,
    interpolation=None,
    force_reencode=False,
    delete_original=False,
    skip_failures=False,
):
    inpath = fou.normalize_path(inpath)
    outpath = fou.normalize_path(outpath)
    in_ext = os.path.splitext(inpath)[1]
    out_ext = os.path.splitext(outpath)[1]

    did_transform = False

    try:
        if (
            size is not None
            or min_size is not None
            or max_size is not None
            or in_ext != out_ext
            or force_reencode
        ):
            img = etai.read(inpath)
            size = _parse_parameters(img, size, min_size, max_size)

        if size is not None:
            if interpolation is not None:
                kwargs = dict(interpolation=interpolation)
            else:
                kwargs = {}

            img = etai.resize(img, width=size[0], height=size[1], **kwargs)
            etai.write(img, outpath)
            did_transform = True
        elif force_reencode or (in_ext != out_ext):
            etai.write(img, outpath)
            did_transform = True
        elif inpath != outpath:
            etau.copy_file(inpath, outpath)

        if delete_original and inpath != outpath:
            etau.delete_file(inpath)
    except Exception as e:
        if not skip_failures:
            raise

        logger.warning(e)

    return did_transform


def _parse_parameters(img, size, min_size, max_size):
    isize = (img.shape[1], img.shape[0])

    if size is not None:
        osize = etai.infer_missing_dims(size, isize)
    else:
        osize = isize

    osize = etai.clip_frame_size(osize, min_size=min_size, max_size=max_size)

    return osize if osize != isize else None


def _parse_ext(ext):
    if ext is None:
        return None

    if not ext.startswith("."):
        ext = "." + ext

    return ext.lower()


def _get_outpath(inpath, output_dir=None, rel_dir=None):
    if output_dir is None:
        return inpath

    if rel_dir is not None:
        rel_dir = fou.normalize_path(rel_dir)
        filename = os.path.relpath(inpath, rel_dir)
    else:
        filename = os.path.basename(inpath)

    return os.path.join(output_dir, filename)
